import pathlib

from lxml import etree
from tomlkit import table

from pnb.mcl.metamodel import standard as metamodel

from dexpi.specificator.formatter import Formatter


def get_default_config():
    config = table()
    config.add('pretty_print', True)
    return config

def make(model_set: metamodel.ModelSet, config: dict):
    out_dir = pathlib.Path(config['out_dir'])
    out_dir.mkdir(parents=True, exist_ok=True)    

    from pnb.mcl.io.xmi import XmiWriter, XmiConfiguration
    out_dir.mkdir(parents=True, exist_ok=True)
    
    from pnb.mcl.metamodel.standard import ModelSet

    #  TODO: file name from config
    etree.ElementTree(XmiWriter(model_set).root).write(
        out_dir / 'Dexpi.xmi', encoding='utf-8', pretty_print=config['pretty_print'])

    xmi_config = XmiConfiguration(primitive_type_prefix='Primitive')
    etree.ElementTree(XmiWriter(model_set, xmi_config).root).write(
        out_dir / 'Dexpi for Modelio.xmi', encoding='utf-8', pretty_print=config['pretty_print'])

FORMATTER = Formatter(
    'xmi', get_default_config, make)



import os
import pathlib
import shutil
import subprocess
import unicodedata

from lxml import etree, html
from sphinx.cmd.build import main as sphinx_build_main

from ...sphinx_ext import set_data, SpecificatorData


def make(model_set, config):
    out_dir = pathlib.Path(config['out_dir'])
    out_dir.mkdir(parents=True, exist_ok=True)  
    
    original_sphinx_project_dir = pathlib.Path(config['sphinx_project'])
    if not original_sphinx_project_dir.is_dir():
        raise Exception() # TODO
    try:
        shutil.rmtree(out_dir)
    except FileNotFoundError:
        pass

    sphinx_project_dir = out_dir / '.sphinx'
    shutil.copytree(original_sphinx_project_dir, sphinx_project_dir)

    (sphinx_project_dir / 'images').mkdir(parents=True, exist_ok=False)
    for path, index in config['index_by_path'].items():
        
        assert path.endswith('.svg')
        path = pathlib.Path(path[:-4] + '.pdf')
        if not path.is_file():
            print("WARN", path)
            
            continue
        shutil.copy(path, sphinx_project_dir / 'images' / f'img{index}.pdf')

    model_dir = sphinx_project_dir / '_models'
    model_dir.mkdir(parents=True)

    data = SpecificatorData(
        model_set=model_set,
        config=config)

    with set_data(data):
        sphinx_build_main(['-M',
            'latex',
            str(sphinx_project_dir),
            str(out_dir),
            '-v'])
        
    clean_latex(out_dir)
    
    tex_path = out_dir / 'latex' / 'dexpispecification.tex'
    
    for _ in range(4):
        subprocess.check_call(
            ['pdflatex',
             'dexpispecification.tex'],
            cwd=out_dir / 'latex')
        
    os.rename(
        out_dir / 'latex' / 'dexpispecification.pdf',
        out_dir / 'DEXPI Specification 2.0.0.pdf')
        

    
def clean_latex(out_dir):

    tex_path = out_dir / 'latex' / 'dexpispecification.tex'
    tex_code = tex_path.read_text(encoding='utf-8') # TODO encoding)
    
    for old, new in [
            (unicodedata.lookup('DOT OPERATOR'), r'$\cdotp$'),
            (unicodedata.lookup('BULLET OPERATOR'), r'$\cdotp$'),
            (unicodedata.lookup('GREEK CAPITAL LETTER THETA'), r'$\Theta$'),
            (r'⁻\(\sp{\text{1}}\)', '$^{-1}$')]:
        tex_code = tex_code.replace(old, new)
    
    lines = []
    for line in tex_code.splitlines(keepends=True):

        use_sphinx_messages = r'\usepackage{sphinxmessages}'
        if line.startswith(use_sphinx_messages):
            line = r'\usepackage{tikz}\usetikzlibrary{arrows,arrows.meta}\usepackage[export]{adjustbox}' + line

        # replace sphinxincludegraphics with input of tex TikZ files.
        sphinx_include = r'\noindent\sphinxincludegraphics{{uml-'
        if line.startswith(sphinx_include):
            line = r'\noindent\adjustbox{max width=\linewidth}{\input{{uml-' + line[len(sphinx_include):] + '}'

        lines.append(line)

    tex_path.write_text(''.join(lines), encoding='utf-8') # TODO encoding?




from unittest import TestCase



import os
from pathlib import Path
import subprocess

from dexpi import specificator


def get_node_bin() -> Path:
    
    env_path = os.environ.get('NODE_BIN')
    if env_path:
        path = Path(env_path)
        if path.is_file():
            return path
        
    path = Path(specificator.__file__).parent / 'bin' / 'node' / 'node.exe'
    if path.is_file():
        return path
    
    return 'node'

        
        
class TestNode(TestCase):
    
    def test_installation(self):
        self.assertTrue(get_node_bin())
        
    def test_call(self):
        version_info = subprocess.run([get_node_bin(), '-v'], capture_output=True).stdout.decode('utf-8')
        self.assertIn(
            'v',
            version_info,
            version_info)



import importlib.metadata
import pathlib

import tomlkit

THIS_DIR = pathlib.Path(__file__).parent

def _get_version() -> str:
    """
    Get the version string of dexpi.specificator.
    """
    if (pyproject_path := THIS_DIR.parent.parent.parent / 'pyproject.toml').is_file():
        with open(pyproject_path, 'rb') as pyproject_file:
            return str(tomlkit.load(pyproject_file)['project']['version'])
    else:
        version = importlib.metadata.version('dexpi.specificator')
        assert isinstance(version, str), type(version)
        return version

__version__ = _get_version()



from collections import namedtuple
import csv
import pathlib
import re
import typing as T

from pnb.mcl.utils import SYMBOL_PATTERN
import ezodf



UomLiteral = namedtuple('UomLiteral', ['row_index',
                                       'rdl_uri',
                                       'rdl_label',
                                       'name',
                                       'un_symbol',
                                       'alt_symbol',
                                       'un_code',
                                       'comment'])


# TODO: move
def cap_case(camel_case: str):
    parts: list[str] = []
    for char in camel_case:
        if char.isupper() and parts:
            parts.append('_')
            parts.append(char)
        else:
            parts.append(char.upper())

    return ''.join(parts)

def dimension_to_rst(dimension: str) -> str:
    if dimension == 1.0:
        return '1'
    for old, new in (
            ('\u207b', '-'),
            ('\u00b9', '1'),
            ('\u00b2', '2'),
            ('\u00b3', '3')):
        dimension = dimension.replace(old, new)

    superscript_chars = '0123456789-'

    # Split in parts that are either supercripted or not.
    parts = re.findall(
        '(?:[{}]+)|(?:[^{}]+)'.format(
            superscript_chars, superscript_chars),
        dimension)

    # TODO: make error invalid dimension
    assert dimension == ''.join(parts)

    fragments = []
    for part in parts:
        if set(part).issubset(superscript_chars):
            fragments.append(f':sup:`{part}`')
        else:
            assert set(part).isdisjoint(superscript_chars)
            fragments.append(part)

    return '\\ '.join(fragments)





def camel_to_normal(text):
    
    # TODO?
    if text == 'pH':
        return text
    
    assert ' ' not in text, text
    result = []
    for char in text:
        if char.isupper():
            if result:
                result.append(' ')
            char = char.lower()
        result.append(char)
    return ''.join(result)

class PhysicalQuantityType:

    def __init__(self, src: str, row_index: int,
                 raw_name: T.Any, physical_dim: T.Any, rdl_uri: T.Any, comment: T.Any):
        self.src = src
        self.row_index = row_index
        self.raw_name = raw_name
        self.physical_dim = physical_dim
        self.rdl_uri = rdl_uri
        self.comment = comment

    def check_args(self, namespace, package):
        ok = True
        if not isinstance(self.raw_name, str):
            ok = False
            namespace.ERROR(self.src, self.row_index,
                            f'{self.raw_name!r} must be a string')
        elif not SYMBOL_PATTERN.fullmatch(self.raw_name):
            ok = False
            namespace.ERROR(self.src, self.row_index,
                            f'{self.raw_name!r} must have pattern of a symbol')
        elif not (isinstance(self.physical_dim, str) or self.physical_dim==1.0):
            ok = False
            namespace.ERROR(self.src, self.row_index,
                            f'{self.physical_dim!r} must be a string or the float 1.0')
        elif not (isinstance(self.rdl_uri, str) or self.rdl_uri is None):
            ok = False
            namespace.ERROR(self.src, self.row_index,
                            f'{self.rdl_uri!r} must be a string or None')
            # TODO: check https://stackoverflow.com/questions/7160737/how-to-validate-a-url-in-python-malformed-or-not
        return ok

    def create(self, namespace, package):
        raise NotImplementedError()

class PhysicalQuantityTypeTemplate(PhysicalQuantityType):

    def __init__(self, uom_literals: list[UomLiteral], description: T.Any, **kwargs):
        super().__init__(**kwargs)
        self.uom_literals = uom_literals
        self.description = description

    def check_args(self, namespace, package):
        ok = super().check_args(namespace, package)
        if not self.uom_literals:
            namespace.WARNING(self.src, self.row_index,
                            f'{self.raw_name!r} has no literals')
        elif not (isinstance(self.description, str) or self.description is None):
            ok = False
            namespace.ERROR(self.src, self.row_index,
                            f'{self.description!r} must be a string or None')

        return ok

    def create(self, namespace, package, s_types=()):
        if not self.check_args(namespace, package):
            return

        # super types not empty for application area types
        # name = f'Nullable{self.raw_name}'
        # abstract_data_type = namespace.ABSTRACT_DATA_TYPE(
        #     superTypes=s_types, description=self.getdoc_for_nullable_type())
        # setattr(package, name, abstract_data_type)
        #
        # name = f'{self.raw_name}'
        # aggregated_data_type = namespace.AGGREGATED_DATA_TYPE(
        #     superTypes=[abstract_data_type], description=self.getdoc_for_type())
        # setattr(package, name, aggregated_data_type)

        name = f'{self.raw_name}Unit'
        
        
        if not s_types:
            s_types = [package.PhysicalQuantityUnit]
            
        enum = namespace.ENUMERATION(
            description=self.getdoc_for_enum(),
            superTypes=s_types)
        setattr(package, name, enum)
        
        for lit in self.uom_literals:
            if not SYMBOL_PATTERN.fullmatch(lit.name):
                #TODO: row index
                namespace.ERROR(self.src, 0,
                                f'{lit.name!r} must have pattern of a symbol')
                continue
            setattr(enum, lit.name, namespace.ENUMERATION_LITERAL(
                rdl_uri=lit.rdl_uri,
                rdl_label=lit.rdl_label,
                un_symbol=lit.un_symbol,
                un_code=lit.un_code))


    def getdoc_for_nullable_type(self):
        raise NotImplementedError()



    def getdoc_for_enum(self):
        # TODO: link for dimension
        return f'''
            A unit of measurement for a physical quantity of type *{camel_to_normal(self.raw_name)}* with dimension
            {dimension_to_rst(self.physical_dim)}.'''

    def getdoc_for_null_value(self):
        REM

class SimplePhysicalQuantityType(PhysicalQuantityTypeTemplate):
    pass


class ApplicationAreaPhysicalQuantityType(PhysicalQuantityTypeTemplate):

    def __init__(self, application_area: str, super_type_name: str, **kwargs):
        super().__init__(**kwargs)
        self.application_area = application_area
        self.super_type_name = super_type_name

    def check_args(self, namespace, package):
        ok = super().check_args(namespace, package)
        if not isinstance(self.application_area, str):
            ok = False
            namespace.ERROR(self.src, self.row_index,
                            f'{self.application_area!r} must be a string')

        return ok

    def getdoc_for_nullable_type(self):
        REM
        

class ApplicationDependentPhysicalQuantityType(PhysicalQuantityType):

    def __init__(self, sub_types: list[ApplicationAreaPhysicalQuantityType], **kwargs):
        super().__init__(**kwargs)
        self.sub_types = sub_types

    def check_args(self, namespace, package):
        ok = super().check_args(namespace, package)
        if not len(self.sub_types) > 1:
            namespace.WARNING(self.src, self.row_index,
                            f'{self.raw_name!r} should have at least two subtypes')

        return ok

    def create(self, namespace, package):
        if not self.check_args(namespace, package):
            return

        name = f'{self.raw_name}Unit'
        
        #print('---', name)
        abstract_data_type = namespace.ABSTRACT_DATA_TYPE(
            description=self.getdoc_for_nullable_type(),
            superTypes=[package.PhysicalQuantityUnit])
        setattr(package, name, abstract_data_type)

        for s_type in self.sub_types:
            s_type.create(namespace, package, [abstract_data_type])

    def getdoc_for_nullable_type(self):
        return f'''
            A unit of measurement for a physical quantity of type *{camel_to_normal(self.raw_name)}* with dimension
            {dimension_to_rst(self.physical_dim)}. <!SELF> has {len(self.sub_types)} subtypes for
            different application areas.'''

        
def read_uom_from_ods(src, namespace, package):
    for type_ in read_ods_uom_table(src, namespace):
        type_.create(namespace, package)

def read_ods_uom_table(src, namespace):
    
    quantity_types = []
    
    doc = ezodf.opendoc(src)
    sheet = doc.sheets[0]
    
    rows = [[item.value for item in ods_row] for ods_row in sheet if len(ods_row)]

    for row_index, row in enumerate(rows, 1):


        # row 1 and 2 contain labels of table
        if row_index < 3:
            continue

        # skip empty lines
        if set(row) == {None}:
            continue

        # for debugging
        #print(str(row_index), row)

        if row[0] is not None:
            raw_name = row[0]
            physical_dim = row[2]
            rdl_uri = row[3]
            comment = row[9]
            description = row[11]
            # reset lists for literals and subtypes
            uom_literals = []
            sub_types = []
            if rows[row_index][1] is None:
                # create simple type
                quantity_types.append(
                    SimplePhysicalQuantityType(
                        src=src,
                        row_index=row_index,
                        raw_name=raw_name,
                        physical_dim=physical_dim,
                        rdl_uri=rdl_uri,
                        uom_literals=uom_literals,
                        comment=comment,
                        description=description))
            else:
                # create application dependent type
                quantity_types.append(
                    ApplicationDependentPhysicalQuantityType(
                        src=src,
                        row_index=row_index,
                        raw_name=raw_name,
                        physical_dim=physical_dim,
                        rdl_uri=rdl_uri,
                        comment=comment,
                        sub_types=sub_types))

            continue

        if row[1] is not None:
            sub_type_name = row[1]
            rdl_uri = row[3]
            comment = row[9]
            application_area = row[10]
            description = row[11]
            # reset list for literals
            uom_literals = []
            # create application area type and update subtypes
            sub_type = ApplicationAreaPhysicalQuantityType(
                src=src,
                row_index=row_index,
                raw_name=sub_type_name,
                physical_dim=physical_dim,
                rdl_uri=rdl_uri,
                uom_literals=uom_literals,
                comment=comment,
                application_area=application_area,
                description=description,
                super_type_name=raw_name)
            sub_types.append(sub_type)

            continue

        # update uom literals
        uom_literals.append(
            UomLiteral(row_index, row[3], row[4], row[5], row[6], row[7], row[8], row[9]))

    return quantity_types


from collections import namedtuple
from pnb.mcl.utils import SYMBOL_PATTERN
import ezodf
import typing as T


EnumerationLiteral = namedtuple('EnumerationLiteral', ['row_index',
                                                       'literal',
                                                       'uri',
                                                       'label',
                                                       'symbol'])



class Enumeration:

    def __init__(self, src: str, row_index: int,
                 name: T.Any, literals: list[EnumerationLiteral]):
        self.src = src
        self.row_index = row_index
        self.name = name
        self.literals = literals
        
    def check_args(self, namespace, package):
        ok = True
        if not isinstance(self.name, str):
            ok = False
            namespace.ERROR(self.src, self.row_index,
                            f'{self.name!r} must be a string')
        elif not SYMBOL_PATTERN.fullmatch(self.name):
            ok = False
            namespace.ERROR(self.src, self.row_index,
                            f'{self.name!r} must have pattern of a symbol')
        elif not self.literals:
            namespace.WARNING(self.src, self.row_index,
                            f'{self.raw_name!r} has no literals')
        return ok

    def create(self, namespace, package):
        if not self.check_args(namespace, package):
            return
        
        enum = namespace.ENUMERATION(description=self.getdoc_for_enum())
        setattr(package, self.name, enum)
        
        for lit in self.literals:
            if not SYMBOL_PATTERN.fullmatch(lit.literal):
                #TODO: row index
                namespace.ERROR(self.src, 0,
                                f'{lit.literal!r} must have pattern of a symbol')
                continue
            setattr(enum, lit.literal, namespace.ENUMERATION_LITERAL(
                rdl_uri=lit.uri,
                rdl_label=lit.label,
                un_symbol=lit.symbol))

    def getdoc_for_enum(self):
        # TODO: doc
        return f'''
            '''



def read_enums_from_ods(src, namespace, package):
    for enum in read_ods_enums_table(src, namespace):
        enum.create(namespace, package)

def read_ods_enums_table(src, namespace):
    
    enums = []
    
    doc = ezodf.opendoc(src)
    sheet = doc.sheets[0]
    
    rows = [[item.value for item in ods_row] for ods_row in sheet if len(ods_row)]

    for row_index, row in enumerate(rows, 1):


        # row 1 contains labels of table
        if row_index < 2:
            continue

        # skip empty lines
        if set(row) == {None}:
            continue

        # for debugging
        #print(str(row_index), row)

        if row[0] is not None:
            name = row[0]
            # reset list for literals
            literals = []
            
            # create enumeration
            enums.append(
                Enumeration(
                    src=src,
                    row_index=row_index,
                    name=name,
                    literals=literals))

            continue

        # update enumeration literals
        literals.append(
            EnumerationLiteral(row_index, row[1], row[2], row[3], row[4]))

    return enums


import os
import pathlib
import shutil

from lxml import etree, html
from sphinx.cmd.build import main as sphinx_build_main

from ...sphinx_ext import set_data, SpecificatorData


def make(model_set, config):
    out_dir = pathlib.Path(config['out_dir'])
    out_dir.mkdir(parents=True, exist_ok=True)  
    
    original_sphinx_project_dir = pathlib.Path(config['sphinx_project'])
    if not original_sphinx_project_dir.is_dir():
        raise Exception() # TODO
    try:
        shutil.rmtree(out_dir)
    except FileNotFoundError:
        pass

    sphinx_project_dir = out_dir / '.sphinx'
    shutil.copytree(original_sphinx_project_dir, sphinx_project_dir)

    (sphinx_project_dir / 'images').mkdir(parents=True, exist_ok=False)
    for path, index in config['index_by_path'].items():
        shutil.copy(path, sphinx_project_dir / 'images' / f'img{index}.svg')

    model_dir = sphinx_project_dir / '_models'
    model_dir.mkdir(parents=True)

    data = SpecificatorData(
        model_set=model_set,
        config=config)

    with set_data(data):
        sphinx_build_main(['-M',
            'html',
            str(sphinx_project_dir),
            str(out_dir),
            '-v'])
            
    clean_html(out_dir)
            
    
def clean_html(out_dir):
    
    print('cleaning HTML...')
    
    for html_path in out_dir.glob('**/*.html'):
        html_tree = html.parse(html_path)
        
        changed = False
        
        for img in html_tree.xpath('//img'):
            # TODO: clean link conversion
            rel_src = img.attrib.get('src')
            if '_images/uml' not in rel_src:
                continue
            changed = True
            abs_src = (html_path.parent / rel_src).resolve()
            assert abs_src.is_file()
            svg_element = etree.parse(abs_src).getroot()

            # Remove SVG namespace.
            for element in svg_element.iter():
                element.tag = element.tag[28:]
            
            del svg_element.attrib['height']
            assert 'style' not in svg_element.attrib
            svg_element.attrib['style'] = 'max-width: 100%; background-color: white;'
            
            for a_element in svg_element.iter('a'):
                href = a_element.attrib['href']
                new_href = str((abs_src.parent / href).relative_to(html_path.parent, walk_up=True)).replace(os.sep, '/')
                new_href = new_href.replace('/_images/../', '/')
                a_element.attrib['href'] = new_href

            img.getparent().replace(img, svg_element)
                
                
           

            
        for admonition_p in html_tree.xpath("//p[@class='admonition-title']"):
            if admonition_p.text:
                admonition_p.text = admonition_p.text.title()
                changed = True

        if changed:
            html_code = etree.tostring(
                html_tree,
                method='html',
                encoding='utf-8')
            
            assert b'<a id="mode_toggle" href="#" :title="mode">' in html_code
            html_code = html_code.replace(
                b'<a id="mode_toggle" href="#" :title="mode">',
                b'<a id="mode_toggle" href="#" @click.prevent="handleClick" :title="mode">')

            html_path.write_bytes(html_code)
            
    print('done')

            
            
            
            
            
            
            
            
            
            
            



from __future__ import annotations

# pylint: disable=missing-docstring, too-few-public-methods

import atexit
import datetime
import inspect
import logging
import json
from pathlib import Path
import re
import textwrap
import traceback
import typing as T

import rdflib

from pnb.mcl.metamodel import standard as metamodel
from pnb.mcl.utils import IDENTIFIER_PATTERN

LOGGER = logging.getLogger(__name__)

DOCUMENTATION_URL = 'https://www.dexpi.org/TODO/'

_NOT_GIVEN = object()


class Source:
    pass


class PySource(Source):

    def __init__(self, path: Path, lineno: int) -> None:
        self.path = path
        self.lineno = lineno

    def __str__(self) -> str:
        
        return f'File "{self.path}", line {self.lineno}'


def py_source_from_calling_frame(extra_frames=0):
    # TODO: doc
    """
    Standard way:
    
        call_info = inspect.getframeinfo(frame)
        return PySource(Path(call_info.filename), call_info.lineno)
    """
    assert extra_frames >= 0
    frame = inspect.currentframe()
    # +1 for the frame calling this function (the frame w.r.t. which we want the calling frame)
    # +1 for the frame of this function
    for _ in range(extra_frames + 2):
        frame = frame.f_back
    filename = inspect.getsourcefile(frame) or inspect.getfile(frame)
    if inspect.istraceback(frame):
        lineno = frame.tb_lineno
    else:
        lineno = frame.f_lineno
    return PySource(Path(filename), lineno)





def context_sort_key(context: ElementInfo | ElementSpecification):
    if isinstance(context, ElementInfo):
        return (1, str(context._source_))
    else:
        assert isinstance(context, ElementSpecification)
        return (2, str(context.qname))



class Message:
    
    level: str
    id: str
    text: str

    def __init__(self, contexts: list[ElementInfo | ElementSpecification], causes: T.Optional[list[Message]]=None) -> None:
        
        self.nr: T.Optional[int] = None
        self.contexts = sorted(contexts, key=context_sort_key)
        if causes:
            causes = sorted(causes, key=lambda msg: msg.nr)
        else:
            causes = []
        self.causes = causes

    def __str__(self) -> str:

        result_rows = [
            f'#{self.nr} --- {self.level}: {self.id} ---']

        for context in self.contexts:
            if isinstance(context, ElementInfo):
                result_rows.append(f'- {context._source_}')
                if context_description := context._get_description_():
                    result_rows.append(f'  {context_description}')
            else:
                result_rows.append(f'- {context.qname}')
                for info in sorted(context.infos, key=context_sort_key):
                    result_rows.append(f'  - {info._source_}')
                    if info_description := info._get_description_():
                        result_rows.append(f'    {info_description}')

        if self.causes:
            causes_repr = ' '.join(f'#{cause.nr}' for cause in self.causes)
            result_rows.append(f'causes: {causes_repr}')

        result_rows += [
            self.text,
            f'(see {DOCUMENTATION_URL}{self.id})']

        return '\n'.join(result_rows)


class FatalErrorMessage(Message):

    level = 'FATAL ERROR'
    id = 'fatal-error'

    def __init__(self, base_traceback: str):
        super().__init__([])
        self.base_traceback = base_traceback

    @property
    def text(self):
        base_traceback_repr = textwrap.indent(self.base_traceback, '  ').rstrip()
        return (
            f'An error has occurred while reading the DSL source files:\n\n{base_traceback_repr}'
            f'\n\nThis error prevents any further processing of the source files.')
            
        
class ErrorMessage(Message):
    
    level = 'ERROR'


class MissingParameter(ErrorMessage):

    id = 'missing-parameter'
    
    def __init__(self, element_infos: list[ElementInfo], parameter_name: str) -> None:
        super().__init__(element_infos)
        self.parameter_name = parameter_name
        
    @property
    def text(self):
        return f'Parameter "{self.parameter_name}" is required, but missing.'
    
    
class InvalidParameterType(ErrorMessage):
    
    id = 'invalid-parameter-type'
    
    def __init__(self,
            element_infos: list[ElementInfo],
            parameter_name: str,
            value: T.Any,
            expected_types: tuple[type, ...]) -> None:
        super().__init__(element_infos)
        self.parameter_name = parameter_name
        self.value = value
        self.expected_types = expected_types
        
    
    @property
    def text(self):
        if len(self.expected_types) == 1:
            expected_types_repr = self.expected_types[0].__name__
        else:
            type_names = ', '.join(type_.__name__ for type_ in self.expected_types)
            expected_types_repr = f'one of {type_names}' 
        return (
            f'Parameter "{self.parameter_name}": got {self.value!r} of type '
            f'{type(self.value).__name__}, but value type must be {expected_types_repr}.')


class InvalidParameter(ErrorMessage):

    parameter_requirements: str

    def __init__(self,
            element_infos: list[ElementInfo],
            parameter_name: str,
            value: T.Any,
            causes: T.Optional[list[Message]]=None):
        super().__init__(element_infos, causes=causes)
        self.parameter_name = parameter_name
        self.value = value

    @property
    def text(self):
        return (
            f'Parameter "{self.parameter_name}": {self.value!r} is no valid value: '
            f'{self.parameter_requirements}')


class NoValidIdentifier(InvalidParameter):

    id = 'no-valid-identifier'
    parameter_requirements = (
        'An identifier must start with [a-zA_Z], followed by an arbitrary number of [a-zA_Z0-9_].')

    def __init__(self,
            element_infos: list[ElementInfo],
            parameter_name: str,
            value: str):
        super().__init__(element_infos, parameter_name, value)


class SkippedElementUsed(InvalidParameter):

    id = 'skipped-element-used'
    parameter_requirements = (
        'It has been skipped.')

    def __init__(self,
            element_infos: list[ElementInfo],
            parameter_name: str,
            value: str,
            causes: list[Message]):
        super().__init__(element_infos, parameter_name, value, causes=causes)


class NoOwner(ErrorMessage):

    id = 'no-owner'

    def __init__(self, element_info: ElementInfo):
        super().__init__([element_info])

    @property
    def text(self):
        return 'The element is skipped because it is not assigned to any owner.'


class OwnerSkipped(ErrorMessage):

    id = 'owner-skipped'

    def __init__(self, element_infos: list[ElementInfo], owner_element_infos: list[ElementInfo]):
        causes = set(info._error_ for info in owner_element_infos)
        causes.discard(None)
        assert causes
        super().__init__(element_infos, causes=list(causes))

    @property
    def text(self):
        return 'The element is skipped because its owner is skipped.'


class MissingMetatype(ErrorMessage):

    id = 'missing-metatype'

    def __init__(self, element_infos: list[ElementInfo]):
        super().__init__(element_infos)

    @property
    def text(self):
        return 'The element is skipped because no metatype is given.'


class MultipleMetatypes(ErrorMessage):

    id = 'multiple-metatypes'

    def __init__(self, element_infos: list[ElementInfo], metatypes: set[T.Type[metamodel.Element]]):
        super().__init__(element_infos)
        self.metatypes = metatypes

    @property
    def text(self):
        metatypes_repr = ', '.join(sorted(type_.__name__ for type_ in self.metatypes))
        return f'The element is skipped because multiple metatypes are given: {metatypes_repr}.'


class InvalidMemberType(ErrorMessage):

    id = 'invalid-member-type'

    def __init__(self,
            element_infos: list[ElementInfo],
            metatype: T.Type[metamodel.Element],
            owner_metatype: T.Type[metamodel.Element]):
        super().__init__(element_infos)
        self.metatype = metatype
        self.owner_metatype = owner_metatype

    @property
    def text(self):
        return (
            f'The element is skipped because its owner of type {self.owner_metatype.__name__} '
            f'cannot have a member of type {self.metatype.__name__}.')


class MultipleParameterValues(ErrorMessage):

    id = 'multiple-parameter-values'

    def __init__(self, element_infos: list[ElementInfo], parameter_name: str):
        super().__init__(element_infos)
        self.parameter_name = parameter_name

    @property
    def text(self):
        return (
            f'The value for parameter "{self.parameter_name}" is given several times. The '
            f'parameter is ignored.')


class CyclicDependency(ErrorMessage):

    id = 'cyclic-dependency'

    def __init__(self, specifications: list[ElementSpecification]):
        super().__init__(specifications)
        self.specifications = specifications

    @property
    def text(self):
        return 'Due to a cyclic dependency, these elements cannot be built.'


# TODO begin ok

class Reference:

    __slots__ = ('_source_', )

    _instances: T.Optional[list[Reference]] = None

    @staticmethod
    def track_instances():
        if Reference._instances is not None:
            raise RuntimeError('instances are already tracked')
        Reference._instances = []

    @staticmethod
    def get_instances() -> list[Reference]:
        if Reference._instances is None:
            raise RuntimeError('instances have not been tracked')
        instances = Reference._instances
        Reference._instances = None
        return instances

    def __init__(self, source: Source):
        if Reference._instances is not None:
            Reference._instances.append(self)
        self._source_ = source

    def __call__(self, **kwargs):
        return ReferenceCall(
            py_source_from_calling_frame(),
            owner=self,
            kwargs=kwargs)

    def __getitem__(self, key: object):
        return ReferenceItem(
            py_source_from_calling_frame(),
            owner=self,
            key=key)

    def __or__(self, other: Reference):
        return ReferenceConjunction(
            py_source_from_calling_frame(),
            first=self,
            second=other)
        
    def _resolve_(self):
        raise NotImplementedError(self)


class ReferenceConjunction(Reference):
    def __init__(self, source: Source, *, first: Reference, second: Reference):
        super().__init__(source)
        self.first = first
        self.second = second
    def _resolve_(self):
        first = self.first._resolve_()
        second = self.second._resolve_()
        for metatype in [metamodel.DataTypeExpression, metamodel.ClassExpression]:
            if isinstance(first, metatype) and isinstance(second, metatype):
                return metamodel.type_conjunction([first, second])
        RAISE


class ReferenceItem(Reference):
    def __init__(self, source: Source, *, owner: Reference, key: object):
        super().__init__(source)
        self.owner = owner
        self.key = key
        
    def _resolve_(self):
        
        binding_dict = self.key
        if not isinstance(binding_dict, dict):
            print(self._source_)
            ji
            raise Postpone() # TODO
        owner = self.owner._resolve_()
        if not isinstance(owner, (metamodel.Class, metamodel.AggregatedDataType)):
            print(vars(self))
            print(self._source_)
            RAISE
            
        if isinstance(owner, metamodel.Class):
            parameter_metatype = metamodel.TypeParameter
            bound_metatype = metamodel.BoundClass
        else:
            parameter_metatype = metamodel.DataTypeParameter
            bound_metatype = metamodel.BoundDataType
             
  
        bindings = []

        
        for parameter_name, type_expression_reference in sorted(binding_dict.items()):
            

            # TODO check parameter_name is name 
            parameter = owner.get(parameter_name)
            if not isinstance(parameter, parameter_metatype):
                print(self._source_)
                print(parameter, parameter_metatype)
                RAISE
            type_expression = type_expression_reference._resolve_()
            # TODO check subtype if not type_expression.
            
            if isinstance(parameter, metamodel.ClassParameter):
                binding_metaclass = metamodel.ClassTemplateParameterBinding
            else:
                binding_metaclass = metamodel.DataTypeTemplateParameterBinding
                
            
                
            
            
            bindings.append(binding_metaclass(parameter=parameter, type=type_expression))
            
        return bound_metatype(owner, bindings)
            



class ReferenceCall(Reference):
    def __init__(self, source: Source, *, owner: Reference, kwargs: dict):
        super().__init__(source)
        self.owner = owner
        self.kwargs = kwargs
        
    def _resolve_(self):
        owner = self.owner._resolve_()
        
        kwargs = {}
        
        
        

        
        #for parameter_name, type_expression_reference in sorted(binding_dict.items()):
            
        #    if parameter_name == 'SingleLanguageStrings':
         #       a=1
            
        
        
        
        
        
        
        
        
        
        
        
        for name, value in self.kwargs.items():
            if isinstance(value, Reference):
                # print('---------', self._source_)
                value = value._resolve_()
                
            elif isinstance(value, list):
                new_value = []
                for v in value:
                    if isinstance(v, Reference):
                        v = v._resolve_()
                    new_value.append(v)
                value = new_value
                        
                
                
                
                
            kwargs[name] = value
        
        if isinstance(owner, (metamodel.AggregatedDataType, metamodel.Class)):

            
            return owner(**kwargs)
            
        
        
        raise Exception(owner, kwargs)

        


class ElementWrapper(Reference):

    def __init__(self, source, *, namespace: WrappedNamespace, name: str):
        super().__init__(source)
        self._namespace = namespace
        self._name = name
    def _resolve_(self):
        value = self._namespace._getter_(self._name)
        if value is None:
            raise Exception(self._name)

        
        return value


# TODO begin ok





class ElementInfo(Reference):
    
    
    __slots__ = ('_macro_name_', '_metatype_', '_name_', '_owner_', '_attributes_', '_qname_',
                 '_error_', '_specification_')

    def __init__(self,
            source: Source, *,
            macro_name: T.Optional[str],
            metatype: T.Optional[T.Type[metamodel.Element]],
            name: T.Optional[str],
            owner: T.Optional[ElementInfo],
            attributes: dict[str, T.Any]) -> None:

        super().__init__(source)
        self._macro_name_ = macro_name
        self._metatype_ = metatype
        self._name_ = name
        self._owner_ = owner
        self._attributes_ = attributes

        self._qname_: T.Optional[str] = None
        self._error_: T.Optional[Message] = None
        self._specification_: T.Optional[ElementSpecification] = None



    #
    # @property
    # def source(self) -> Source:
    #     return self._source_
    #


    def _get_description_(self) -> str:
        if self._macro_name_:
            if self._name_:
                return f'{self._macro_name_} "{self._name_}"'
            else:
                return self._macro_name_
        else:
            if self._name_:
                return f'reference to "{self._name_}"'
            else:
                return f'reference'

    def _set_name_and_owner(self, name: str, owner: ElementInfo):
        if self._name_ is not None:
            RAISEERROR
        if self._owner_ is not None:
            RAISEERROR
        self._name_ = name
        self._owner_ = owner
    
    def __setattr__(self, name, value):
        if name.startswith('_'):
            calling_frame = inspect.currentframe().f_back
            calling_filename = inspect.getsourcefile(calling_frame)
            if calling_filename != __file__:
                TODO
            super(). __setattr__(name, value)
        elif isinstance(value, ElementInfo):
            value._set_name_and_owner(name, self)
        else:
            raise TypeError(f'only ElementInfos can be attributes of an ElementInfo; got {value!r}') 
    
    
    # TODO begin ok
    
    def __getattr__(self, name: str) -> ElementInfo:
        
        return ElementInfo(
            source=py_source_from_calling_frame(),
            macro_name=None,
            metatype=None,
            name=name,
            owner=self,
            attributes={})
    
    # TODO end ok

        



    def __format__(self, fmt):
        
        if not fmt.startswith('sp'):
            return super().__format__(fmt)
        fmt = fmt[2:]
        assert fmt[0] and fmt[0] in '-!~'
        if fmt[0] == '-':
            fmt = fmt[1:]
        return f'<SPREF-{id(self)}-{fmt}>'
    
    def _resolve_(self):
        specification = self._specification_
        if not specification:
            print(vars(self))
            RAISE
        return specification.resolve()
    
    
    
    


    
 

class TemplateModel(metamodel.Namespace):
    
    __slots__ = []
    
    ownedTemplates = (
        metamodel.MutableMembersProperty[metamodel.Property](
            __slots__, metamodel.Property, True,
            derived_properties=[metamodel.Namespace.ownedMembers]))
    
    def add(self, template):
        if isinstance(template, metamodel.Property):
            self.ownedTemplates.add(template)
        else:
            raise TypeError(template)


class ElementMacro:
    
    __slots__ = ('_name', '_metatype')

    def __init__(self, name: str, metatype: T.Type[metamodel.Element]) -> None:
        self._name = name
        self._metatype = metatype

    def __call__(self, **kwargs) -> ElementInfo:
        return ElementInfo(
            source=py_source_from_calling_frame(),
            macro_name=self._name,
            metatype=self._metatype,
            name=kwargs.pop('name', None),
            owner=None,
            attributes=kwargs)


def _make_macro_by_name():
    macro_by_name = {}
    for metamodel_type in [
            metamodel.AbstractDataType,
            metamodel.DataProperty,
            metamodel.Enumeration,
            metamodel.EnumerationLiteral,
            metamodel.Model,
            metamodel.Package,
            metamodel.AbstractClass,
            metamodel.ConcreteClass,
            metamodel.ReferenceProperty,
            metamodel.AggregatedDataType,
            metamodel.CompositionProperty,
            metamodel.DataTypeParameter,
            metamodel.ClassParameter,
            TemplateModel]:

        # E.g.: ConcreteClass -> "CONCRETE_CLASS"
        name = '_'.join(
            word.upper() for word in re.findall('[A-Z][^A-Z]*', metamodel_type.__name__))
        macro_by_name[name] = ElementMacro(name, metamodel_type)
    return macro_by_name

MACRO_BY_NAME = _make_macro_by_name()



class Rdl2Namespace:
    
    def __getattr__(self, name):
        return JordName(name.replace('_', ' '))
    
    def term(self, name):
        return Rdl2Name(name)
    
    
class Rdl2Name(str):
    
    def get_uri_label_and_description(self, rdl_data):
        
       # rdl_data = rdl_data.get('RDL
        
        result = rdl_data.get(str(self))
        if result:
            return result

        query_code = f'''
            SELECT ?uri ?definition
            WHERE {{
              SERVICE <https://data.posccaesar.org/rdl/sparql> {{
                ?uri <http://www.w3.org/2000/01/rdf-schema#label> {rdflib.Literal(self).n3()} .
                OPTIONAL {{ ?uri <http://data.posccaesar.org/rdl/hasDefinition> ?definition }}
              }}
            }}'''
        
        rows = list(rdflib.Graph().query(query_code))
        if not rows:
            raise Exception(label)
        elif len(rows) > 1:
            raise Exception(label, rows)
        else:
            uri, definition = rows[0]
            uri = str(uri)
            definition = str(definition)
          
        result = uri, str(self), str(definition)
        rdl_data[str(self)] = result

        return result

class JordNamespace:
    
    def __getattr__(self, name):
        return JordName(name.replace('_', ' '))
    
    def term(self, name):
        return JordName(name)

class JordName(str):
    
    def get_uri_label_and_description(self, all_rdl_data):
        
        rdl_data = all_rdl_data.get('jord')
        if rdl_data is None:
            rdl_data = {}
            all_rdl_data['jord'] = rdl_data

        result = rdl_data.get(str(self))
        if result:
            return result

        query_code = f'''
            SELECT ?uri ?definition
            WHERE {{
              SERVICE <https://data.posccaesar.org/rdl/sparql> {{
                ?uri <http://www.w3.org/2000/01/rdf-schema#label> {rdflib.Literal(self).n3()} .
                OPTIONAL {{ ?uri <http://data.posccaesar.org/rdl/hasDefinition> ?definition }}
              }}
            }}'''
        
        rows = list(rdflib.Graph().query(query_code))
        if not rows:
            raise Exception(label)
        elif len(rows) > 1:
            raise Exception(label, rows)
        else:
            uri, definition = rows[0]
            uri = str(uri)
            definition = str(definition)
          
        result = uri, str(self), str(definition)
        rdl_data[str(self)] = result

        return result


class DexpiNamespace:
    
    def __getattr__(self, name):
        return DexpiName(name.replace('_', ' '))
    
    def term(self, name):
        return DexpiName(name)


class DexpiName(str):
    
    def get_uri_label_and_description(self, rdl_data):
        namespace = 'http://sandbox.dexpi.org/rdl/'
        return namespace + cap_to_camel_case(self), str(self), None














class TODO_MACRO:
    def __getattr__(self, name):
        pass
    def term(self, name):
        pass
    




AUTO = object()






class WrappedNamespace:
    
    def __init__(self, name: str, getter):
        
        self._name_ = name
        self._getter_ = getter
        
    
    def __getattr__(self, name):
        return ElementWrapper(
            py_source_from_calling_frame(),
            namespace=self,
            name=name)
        
        

BUILTIN = WrappedNamespace('BUILTIN', {
    'AnyURI': metamodel.BUILTIN.AnyURI,
    'Boolean': metamodel.BUILTIN.Boolean,
    'DateTime': metamodel.BUILTIN.DateTime,
    'Double': metamodel.BUILTIN.Double,
    'Integer': metamodel.BUILTIN.Integer,
    'String': metamodel.BUILTIN.String,
    'Undefined': metamodel.BUILTIN.Undefined,
    'UnsignedByte': metamodel.BUILTIN.UnsignedByte}.get)










MACRO_BY_NAME.update(dict(
  
  
    BUILTIN = BUILTIN,

    AUTO=AUTO,
    TODO=print,
    WARNING=print,
    ERROR=print,
    FATAL_ERROR=print,
    JORD_RDL=JordNamespace(),
    RDL2=Rdl2Namespace(),
    DEXPI_RDL=DexpiNamespace()))


MEMBER_METATYPES_BY_METATYPE: dict[T.Type[metamodel.Element], tuple[T.Type[metamodel.NamedElement], ...]] = {
    metamodel.AbstractClass: (metamodel.Property, metamodel.TypeParameter),
    metamodel.AbstractDataType: (metamodel.AbstractDataType, ),
    metamodel.AggregatedDataType: (metamodel.DataProperty, metamodel.DataTypeParameter),
    metamodel.ConcreteClass: (metamodel.Property, metamodel.TypeParameter),
    metamodel.Enumeration: (metamodel.Type, metamodel.EnumerationLiteral),
    metamodel.Model: (metamodel.Type, metamodel.Package, metamodel.SingletonValue),
    metamodel.Package: (metamodel.Type, metamodel.Package, metamodel.SingletonValue),
    TemplateModel: (metamodel.Property, )}


class TempTemplate:
    
    def __getattr__(self, name):
        return name


TEMP_TEMPLATE = lambda: TempTemplate()


class SpNamespace:
    
    def __init__(self, members: dict):
        self.__dict__.update(members)

    def __getitem__(self, name):
        return self.__dict__[name]
    
    def __setitem__(self, name, value):
        if 0 and TODO and (isinstance(value, ElementInfo) and value._metatype_ is metamodel.Model
                or isinstance(value, Templates)):
            if name in self.__dict__:
                ERROR
            value._set_name_and_owner(name, None)
        self.__dict__[name] = value
    
    def __iter__(self):
        return iter(self.__dict__)
    
    def __len__(self):
        return len(self.__dict__)


class DslReader:

    def __init__(self,
            src: str | Path | T.Iterable[str|Path],
            cache_dir: T.Optional[Path]=None) -> None:

        # TOLATERDO: move to separate util
        # TOLATERDO: make class rather than instance attribute
        rdl_data = {}
        if cache_dir:
            rdl_data_path = cache_dir / 'rdl_data.json'
            if rdl_data_path.exists():
                # pylint: disable=broad-exception-caught
                try:
                    with rdl_data_path.open('r') as fin:
                        rdl_data = json.load(fin)
                except Exception as error:
                    print(error) # TODO log

            def save(rdl_data=rdl_data, rdl_data_path=rdl_data_path):
                # pylint: disable=dangerous-default-value
                # pylint: disable=broad-exception-caught
                try:
                    rdl_data_path.parent.mkdir(parents=True, exist_ok=True)
                    with rdl_data_path.open('w') as fout:
                        json.dump(rdl_data, fout, sort_keys=True, indent=4)
                except Exception as error:
                    print(error) # TODO log
            atexit.register(save)
        self.rdl_data = rdl_data

        self.messages: list[Message] = []

        # TODO: remove
        self.model_by_name: dict[str, metamodel.Model] = {
            'Builtin': metamodel.BUILTIN}
        self.templates_by_name: dict[str, TemplateModel] = {}
        self.metadata = metamodel.MetaData(
            name='MetaData',
            uri='https://data.dexpi.org/models/2.0.0/MetaData.xml')
        self.index_by_path = {}

        LOGGER.info('read sources')
        references = self._read_sources(src)
        LOGGER.info('build models')
        element_infos = [info for info in references if isinstance(info, ElementInfo)]
        element_infos = self._element_infos_with_qname(element_infos)
        element_specifications = self._build_element_specifications(element_infos)
        self._build_elements(element_specifications)
        LOGGER.info('%s models built:', len(self.model_by_name)-1)
        for name in sorted(self.model_by_name):
            if name != 'Builtin':
                LOGGER.info('- %s', name)
                
        self.model_set = metamodel.ModelSet()
        for model in self.model_by_name.values():
            self.model_set.add(model)
            
        self.model_set.add(self.metadata)
            

    def _add_message(self, message: Message) -> None:
        assert message.nr is None
        self.messages.append(message)
        message.nr = len(self.messages)
        LOGGER.warning(message)

    def _iter_src(self, src: str | Path | T.Iterable[str|Path]) -> T.Iterator[str|Path]:
        if isinstance(src, (str, Path)):
            src = [src]
        for single_src in src:
            if isinstance(single_src, str):
                yield single_src
            elif single_src.is_file():
                yield single_src
            elif single_src.is_dir():
                yield from single_src.rglob('*.py')

    def _read_sources(self, src: str | Path | T.Iterable[str|Path]) -> list[Reference]:

        Reference.track_instances()

        try:
            for single_src in self._iter_src(src):
                LOGGER.info('- read %s', single_src)
                local_vars = SpNamespace(MACRO_BY_NAME)
                local_vars.NAMESPACE = local_vars
                if isinstance(single_src, Path):
                    code = compile(single_src.read_bytes(), str(single_src), 'exec')
                    local_vars.__file__ = str(single_src)
                    local_vars.THIS_DIR = single_src.parent
                else:
                    code = compile(single_src, '<string>', 'exec')
                exec(code, None, local_vars)
        except Exception as error:
            traceback_lines = traceback.format_exc().splitlines()
            del traceback_lines[1]
            del traceback_lines[1]
            msg = FatalErrorMessage('\n'.join(traceback_lines))
            self._add_message(msg)
            Reference.get_instances()
            return []
        else:
            return Reference.get_instances()


    def _element_infos_with_qname(self, infos: list[ElementInfo]) -> list[ElementInfo]:

        msg: ErrorMessage

        infos_to_handle: list[ElementInfo] = []
        for info in infos:
            name = info._name_
            if not isinstance(name, (str, type(None))):
                msg = InvalidParameterType([info], 'name', name, (str, type(None)))
                info._error_ = msg
                self._add_message(msg)
                continue
            if isinstance(name, str) and not IDENTIFIER_PATTERN.match(name):
                msg = NoValidIdentifier([info], 'name', name)
                info._error_ = msg
                self._add_message(msg)
                continue
            if name is None and (
                    info._owner_ or info._metatype_ in (metamodel.Model, TemplateModel)):
                # Case 'no owner' is handled below.
                msg = MissingParameter([info], 'name')
                info._error_ = msg
                self._add_message(msg)
                continue
            infos_to_handle.append(info)

        progress_made = True
        infos_with_qname: list[ElementInfo] = []
        while progress_made:
            progress_made = False
            infos = infos_to_handle
            infos_to_handle = []

            for info in infos:
                owner = info._owner_
                if owner is None:
                    if info._metatype_ in (metamodel.Model, TemplateModel):
                        info._qname_ = info._name_
                        infos_with_qname.append(info)
                    else:
                        msg = NoOwner(info)
                        info._error_ = msg
                        self._add_message(msg)
                    progress_made = True
                elif owner._error_:
                    msg = OwnerSkipped([info], [owner])
                    info._error_ = msg
                    self._add_message(msg)
                    progress_made = True
                elif owner._qname_:
                    info._qname_ = owner._qname_ + f'.{info._name_}'
                    infos_with_qname.append(info)
                else:
                    infos_to_handle.append(info)
            infos = infos_to_handle

        assert not infos_to_handle

        return infos_with_qname


    def _build_element_specifications(self, infos: list[ElementInfo]) -> list[ElementSpecification]:

        infos_by_qname: dict[str, list[ElementInfo]] = {}
        for info in infos:
            assert info._qname_
            infos_by_qname.setdefault(info._qname_, []).append(info)

        specifications: list[ElementSpecification] = []
        for qname, infos_with_qname in sorted(infos_by_qname.items()):
            specification = self._build_element_specification(qname, infos_with_qname)
            if specification:
                specifications.append(specification)

        return specifications


    def _build_element_specification(self,
            qname: str, infos: list[ElementInfo]) -> T.Optional[ElementSpecification]:

        msg: ErrorMessage

        metatypes = set()
        for info in infos:
            if info._metatype_:
                metatypes.add(info._metatype_)
        if not metatypes:
            msg = MissingMetatype(infos)
            for info in infos:
                info._error_ = msg
            self._add_message(msg)
            return None
        if len(metatypes) > 1:
            msg = MultipleMetatypes([info for info in infos if info._metatype_], metatypes)
            for info in infos:
                info._error_ = msg
            self._add_message(msg)
            return None

        metatype = metatypes.pop()
        metatype_infos = [info for info in infos if info._metatype_]

        owner_info = infos[0]._owner_
        owner_specification: T.Optional[ElementSpecification] = None
        # TOLATERDO: Do why need to check anything for Model, TemplateModel?
        if metatype not in (metamodel.Model, TemplateModel):
            assert owner_info
            owner_specification = owner_info._specification_
            if owner_specification:
                if not issubclass(
                        metatype,
                        MEMBER_METATYPES_BY_METATYPE.get(owner_specification.metatype, ())):
                    msg = InvalidMemberType(infos, metatype, owner_specification.metatype)
                    for info in infos:
                        info._error_ = msg
                    self._add_message(msg)
                    return None
            else:
                owner_infos = []
                for info in infos:
                    assert info._owner_
                    owner_infos.append(info._owner_)
                msg = OwnerSkipped(infos, owner_infos)
                for info in infos:
                    info._error_ = msg
                self._add_message(msg)
                return None

        infos_by_attribute_name: dict[str, list[ElementInfo]] = {}
        for info in infos:
            for attribute_name in info._attributes_:
                infos_by_attribute_name.setdefault(attribute_name, []).append(info)

        value_by_attribute_name: dict[str, T.Any] = {}
        error_by_attribute_name: dict[str, Message] = {}
        for attribute_name, attribute_infos in infos_by_attribute_name.items():
            if len(attribute_infos) > 1:
                msg = MultipleParameterValues(attribute_infos, attribute_name)
                error_by_attribute_name[attribute_name] = msg
                self._add_message(msg)
            else:
                value_by_attribute_name[attribute_name] = (
                    attribute_infos[0]._attributes_[attribute_name])

        specification = ElementSpecification(
            qname=qname,
            infos=infos,
            owner_specification=owner_specification,
            metatype=metatype,
            metatype_infos=metatype_infos,
            value_by_attribute_name=value_by_attribute_name,
            infos_by_attribute_name=infos_by_attribute_name,
            error_by_attribute_name=error_by_attribute_name)
        for info in infos:
            info._specification_ = specification
        if owner_specification:
            owner_specification.owned_specifications.append(specification)

        return specification


    def _build_elements(self, element_specifications: list[ElementSpecification]) -> None:
        
        msg: ErrorMessage
        
        self._metadata_cache = []

        type_specifications: list[ElementSpecification] = []
        specifications_by_priority: dict[int, list[ElementSpecification]] = {}

        for spec in element_specifications:
            specifications_by_priority.setdefault(spec.get_priority(), []).append(spec)
            if issubclass(spec.metatype, metamodel.Type):
                type_specifications.append(spec)

        types_and_template_usages = []

        for priority, specifications in sorted(specifications_by_priority.items()):

            progress_made = True
            while progress_made:
                progress_made = False
                to_build: list[ElementSpecification] = []
                for spec in specifications:
                    try:
                        element = spec.build(self)
                        assert element, spec.metatype
                        progress_made = True
                        if isinstance(element, metamodel.Model):
                            assert element.name not in self.model_by_name
                            self.model_by_name[element.name] = element
                        elif isinstance(element, TemplateModel):
                            assert element.name not in self.templates_by_name
                            self.templates_by_name[element.name] = element
                        self._add_metadata(element, spec.value_by_attribute_name, 
                                           spec.infos_by_attribute_name, spec.metatype_infos[0]._source_)
                    except Postpone:
                        to_build.append(spec)
                specifications = to_build
                
            if specifications:
                msg = CyclicDependency(specifications)
                for specification in specifications:
                    assert specification.element is None
                    assert specification.error is None
                    specification.error = msg
                self._add_message(msg)

        for spec in type_specifications:
            type_element = spec.element
            if not type_element:
                continue

            if (template_usages := spec.value_by_attribute_name.get('templates')) is not None:

                if not isinstance(template_usages, list):
                    msg = InvalidParameterType(
                        spec.infos_by_attribute_name['templates'],
                        'templates',
                        template_usages,
                        (list, ))
                    self._add_message(msg)
                    continue

                for usage_nr, usage in enumerate(template_usages):
                    if not isinstance(usage, ElementInfo):
                        msg = InvalidParameterType(
                            spec.infos_by_attribute_name['templates'],
                            f'templates[{usage_nr}]',
                            usage,
                            (metamodel.Property, ))
                        self._add_message(msg)
                        continue

                    if usage._error_:
                        msg = SkippedElementUsed(
                            spec.infos_by_attribute_name['templates'],
                            f'templates[{usage_nr}]',
                            usage._qname_,
                            causes=[usage._error_])
                        self._add_message(msg)
                        continue

                    assert usage._specification_
                    prop = usage._specification_.element
                    if not isinstance(prop, metamodel.Property):
                        msg = InvalidParameterType(
                            spec.infos_by_attribute_name['templates'],
                            f'templates[{usage_nr}]',
                            prop,
                            (metamodel.Property, ))
                        self._add_message(msg)
                        continue

                    prop_kwargs = {
                        'name': prop.name,
                        'lower': prop.lower,
                        'upper': prop.upper,
                        'type_': prop.type,
                        'isOrdered': prop.isOrdered}
                    
                    if isinstance(prop, metamodel.DataProperty):
                        prop_kwargs['isUnique'] = prop.isUnique
                    elif isinstance(prop, metamodel.ReferenceProperty):
                        prop_kwargs['isUnique'] = prop.isUnique
                        prop_kwargs['oppositeLower'] = prop.oppositeLower
                        prop_kwargs['oppositeUpper'] = prop.oppositeUpper
                    new_prop = type(prop)(**prop_kwargs)
                    type_element.add(new_prop)
                    self._add_metadata(new_prop, usage._specification_.value_by_attribute_name,
                                       usage._specification_.infos_by_attribute_name, usage._specification_.metatype_infos[0]._source_)

    
        for args in self._metadata_cache:
            assert args[0]
            self._do_add_metadata(*args)

        
    def _add_metadata(self, element, value_by_attribute_name, infos_by_attribute_name, source):
        
        self._metadata_cache.append((element, value_by_attribute_name, infos_by_attribute_name, source))
        
        
    def _do_add_metadata(self, element, value_by_attribute_name, infos_by_attribute_name, source):
        
        

        #doc = self.value_by_attribute_name.get('doc')
        #if isinstance(doc, str):
        #    metadata.set(element, 'doc', doc)
            
        import textwrap
        
            
        for name in ['rdl_uri', 'rdl_label', 'un_symbol', 'un_code', 'description', 'doc', 'details']:
            value = value_by_attribute_name.get(name)
            

            if isinstance(value, str):
                
                if name in ['details', 'description', 'doc']:
                    
                    
                    if name == 'doc':
                        assert 'description' not in value_by_attribute_name, element
                        name = 'description'
                    
                    value = textwrap.dedent(value).strip()
                    
                    if 'SELF' in value:
                        assert isinstance(element, (metamodel.Type, metamodel.Package, metamodel.Model)), element.qualifiedName
                        def on_self(matchobj):
                            nolink = matchobj.group('nolink')
                            path = matchobj.group('path')
                            return f':sp:element:`{nolink}~{element.qualifiedName}{path}`'
                        value = re.sub(
                            r'\<(?P<nolink>\!?)SELF(?P<path>[^\>]*)\>',
                            on_self,
                            value)
                        
                    if 'OWNER' in value:
                        assert isinstance(element, (metamodel.Property, metamodel.TypeParameter)), element.qualifiedName
                        def on_owner(matchobj):
                            nolink = matchobj.group('nolink')
                            path = matchobj.group('path')
                            return f':sp:element:`{nolink}~{element.owner.qualifiedName}{path}`'
                        value = re.sub(
                            r'\<(?P<nolink>\!?)OWNER(?P<path>[^\>]*)\>',
                            on_owner,
                            value)
                        
                    if 'SPREF' in value:
                        ko
                        
                        # TODO: still used?

                        def on_spref(matchobj):
                            info_nr = matchobj.group('info_nr')
                            modifier = matchobj.group('modifier')
                            path = matchobj.group('path')

                            assert modifier in ('', '!', '~', '!~', '~!')
                            if path:
                                assert path[0] != '.'
                                path = '.' + path
                                
                                
                            base_info = element_info_by_id[info_nr]
                            base_path = base_info._qname_ # TODO handle None
                            assert base_path

                            return f':sp:element:`{modifier}{base_path}{path}`'

                        value = re.sub(
                            r'\<SPREF-(?P<info_nr>[0-9]*)-(?P<modifier>[\~\!]*)(?P<path>[^\>]*)\>',
                            on_spref,
                            value)
                        
                        
                    if 'image' in value:
                     #   .. image:: example1.*
                     
                      

                        
                        
                        def on_image(matchobj):
                            
                            src_infos = infos_by_attribute_name.get(name)
                            assert len(src_infos) == 1
                        
                            
                            
                            src_dir = src_infos[0]._source_.path.parent
     
                            file_name = matchobj.group('file_name')
                            if file_name.endswith('.*'):
                                file_name = file_name[:-2] + '.svg'
                            file_path = src_dir / file_name

                            if not file_path.is_file():
                                return '???' # TODO

                            file_path = str(file_path)
                            
                            index = self.index_by_path.get(file_path)
                            if index is None:
                                index = len(self.index_by_path)
                                self.index_by_path[file_path] = index
                                
                            return f'.. image:: /images/img{index}.*'


                        value = re.sub(
                            r'\.\. image:: (?P<file_name>.*?)$',
                            on_image,
                            value,
                            flags=re.MULTILINE)
       

                self.metadata.set(element, name, value)
                
                
        
        rdl = value_by_attribute_name.get('rdl')
        if rdl is not None:
            rdl_uri, rdl_label, rdl_definition = rdl.get_uri_label_and_description(self.rdl_data)
            assert self.metadata.get(element, 'rdl_uri', None) is None
            assert self.metadata.get(element, 'rdl_label', None) is None
            self.metadata.set(element, 'rdl_uri', rdl_uri)
            self.metadata.set(element, 'rdl_label', rdl_label)
            
            
            if value_by_attribute_name.get('description') is AUTO:
                self.metadata.set(element, 'description', rdl_definition)
                self.metadata.set(element, 'description_from_rdl', True)
        else:
            if value_by_attribute_name.get('description') is AUTO:
                print(element)
                # TODO ERROR
            
                
        exampleValue = value_by_attribute_name.get('exampleValue')
        if exampleValue is not None:
            if isinstance(exampleValue, (str, datetime.datetime, int, float)):
                nr_of_values = 1
            
            elif isinstance(exampleValue, Reference):
                exampleValue = exampleValue._resolve_()
                assert isinstance(exampleValue, (metamodel.Object, metamodel.AggregatedDataValue, metamodel.EnumerationLiteral)), exampleValue
                nr_of_values = 1
            elif isinstance(exampleValue, list):
                tempValue = []
                for item in exampleValue:
                    if isinstance(item, (str, datetime.datetime, int, float)):
                        pass
                    elif isinstance(item, Reference):
                        item = item._resolve_()
                        assert isinstance(item, (metamodel.Object, metamodel.AggregatedDataValue, metamodel.EnumerationLiteral)), exampleValue
                    tempValue.append(item)
                exampleValue = tempValue
                nr_of_values = len(exampleValue)
            else:
                raise Exception(exampleValue)
            
            
            if nr_of_values < element.lower:
                print('###', source)
                print('### need', element.lower, 'values for ', element.qualifiedName)
                exampleValue = None
            if element.upper is not None and element.upper < nr_of_values:
                print('###', source)
                print('### need at most', element.upper, 'values for ', element.qualifiedName)
                exampleValue = None
                
            if nr_of_values == 1 and (element.upper is None or element.upper > 1) and not isinstance(exampleValue, list):
                exampleValue = [exampleValue]

                
                #print('###', source)
                #print('### need a list for ', element.qualifiedName)
                
                
            if exampleValue is not None:
                self.metadata.set(element, 'exampleValue', exampleValue)
   
            
        elif isinstance(element, metamodel.DataProperty):
            
            if 'Process' not in element.qualifiedName:
            
                print('###', source)
                print('### no example property for', element.qualifiedName)



class Postpone(Exception):
    pass


class ElementSpecification:
    def __init__(self,
            qname: str,
            infos: list[ElementInfo],
            owner_specification: T.Optional[ElementSpecification],
            metatype: T.Type[metamodel.Element],
            metatype_infos: list[ElementInfo],
            value_by_attribute_name: dict[str, T.Any],
            infos_by_attribute_name: dict[str, list[ElementInfo]],
            error_by_attribute_name: dict[str, Message]):
        
        self.name = qname.rsplit('.', 1)[-1]
        self.qname = qname
        self.infos = infos
        self.owner_specification = owner_specification
        self.metatype = metatype
        self.metatype_infos = metatype_infos
        self.owned_specifications: list[ElementSpecification] = []
        self.value_by_attribute_name = value_by_attribute_name
        self.infos_by_attribute_name = infos_by_attribute_name
        self.error_by_attribute_name = error_by_attribute_name
        self.element: T.Optional[metamodel.NamedElement] = None
        self.error: T.Optional[Message] = None

        
    def resolve(self):
        if self.error:
            RAISE
        element = self.element
        if not self.element:
            print('#########', self.qname)
            raise Postpone()
        return element
        
        
        
        
    property_priority = 6
        
        
    def get_priority(self) -> int:
        for priority, metatype in enumerate([
                metamodel.Model,
                metamodel.Package,
                metamodel.Type,
                metamodel.TypeParameter,
                TemplateModel,
                metamodel.EnumerationLiteral,
                metamodel.Property]):

            if issubclass(self.metatype, metatype):
                return priority
        raise Exception(self.metatype)
    
    
    
    def _get_required_attribute(self, attribute_name: str, reader: DslReader):
        value = self.value_by_attribute_name.get(attribute_name, _NOT_GIVEN)
        if value is _NOT_GIVEN:
            error_msg = self.error_by_attribute_name.get(attribute_name)
            if error_msg:
                COV
            else:
                COV
        return value
    
    
    
    def _get_required_uri_attribute(self, attribute_name: str, reader: DslReader):
        
        uri = self._get_required_attribute(attribute_name, reader)
        if not isinstance(uri, str):
            COV
        # TOLATERDO: check pattern?
        return uri
        
        
    def _get_type_attributes(self, attribute_name):
        
        

       
        
        if attribute_name in self.error_by_attribute_name:
            return []  # TODO
        type_specifications = self.value_by_attribute_name.get(attribute_name, [])
        types = [ts._specification_.element for ts in type_specifications]
        if None in types:
            raise Postpone()
        return types
    
    
    def _get_non_negative_integer_attribute(self, attribute_name):
        value = self.value_by_attribute_name.get(attribute_name)
        assert isinstance(value, int)
        assert value >= 0
        return value
    
    def _get_optional_non_negative_integer_attribute(self, attribute_name):
        value = self.value_by_attribute_name.get(attribute_name)
        assert value is None or isinstance(value, int)
        assert value is None or value >= 0
        return value
    
    def _get_optional_bool_attribute(self, attribute_name):
        value = self.value_by_attribute_name.get(attribute_name)
        assert value is None or isinstance(value, bool)
        return value

    
    
    
    
    
    
    
    
    def _get_type_expression_attribute(self, attribute_name):
        if attribute_name in self.error_by_attribute_name:
            RAISE
            
        type_expression_reference = self.value_by_attribute_name.get(attribute_name, None)
        
        if type_expression_reference is None:
            RAISE
            
        if not isinstance(type_expression_reference, Reference):
            RAISE
            
        
            
        type_expression = type_expression_reference._resolve_()
        
        #if not isinstance(type_expression, metamodel.TypeExpression):
        #    RAISE
        
        
        return type_expression

            
        

        
        
        
    def get_owner_element(self, reader: DslReader) -> T.Optional[ElementSpecification]:
        owner_specification = self.owner_specification
        if self.metatype in (metamodel.Model, TemplateModel):
            assert not owner_specification
            return None
        assert owner_specification
        if owner_specification.error:
            assert owner_specification.element is None
            RAISE
        return owner_specification.element
        
        
        
        
    def build(self, reader: DslReader):
        
        owner_element = self.get_owner_element(reader)
        
        if self.metatype is metamodel.Model:
            assert not owner_element
            build_element = self._build_model
        elif self.metatype is TemplateModel:
            assert not owner_element
            build_element = self._build_templates
        elif not owner_element:
            build_element = None
        elif self.metatype is metamodel.Package:
            build_element = self._build_package
        elif issubclass(self.metatype, metamodel.Type):
            build_element = self._build_type
        elif issubclass(self.metatype, metamodel.TypeParameter):
            build_element = self._build_type_parameter
        elif issubclass(self.metatype, metamodel.Property):
            build_element = self._build_property
        elif issubclass(self.metatype, metamodel.EnumerationLiteral):
            build_element = self._build_enumeration_literal
        else:
            return None and TODORAISE
            raise Exception(self.metatype)
              
        if build_element:
            element = build_element(reader)
            if element:
                self.element = element
                if owner_element:
                    # E.g., EnumerationLiterals are added to their Enumerations during construction.
                    if element.owner is not owner_element:
                        owner_element.add(element)
        else:
            element = None
                
        return element

    def _build_model(self, reader: DslReader) -> metamodel.Model:
        uri = self._get_required_uri_attribute('uri', reader)
        return metamodel.Model(name=self.name, uri=uri)
    
    def _build_templates(self, reader: DslReader) -> TemplateModel:
        return TemplateModel(name=self.name)
        
    def _build_package(self, reader: DslReader) -> metamodel.Package:
        return metamodel.Package(name=self.name)
    
    def _build_enumeration_literal(self, reader) -> metamodel.EnumerationLiteral:
        owner_element = self.get_owner_element(reader)
        assert isinstance(owner_element, metamodel.Enumeration)
        # TODO: how to keep order?
        return metamodel.EnumerationLiteral(name=self.name, enumeration=owner_element)
    
    def _build_type(self, reader: DslReader):
 
        super_types = self._get_type_attributes('superTypes')
        
        metatype = self.metatype
        if issubclass(metatype, metamodel.Class):
            super_type_metatype = metamodel.Class
        else:
            assert issubclass(metatype, metamodel.DataType)
            super_type_metatype = metamodel.AbstractDataType
            
        invalid_super_types = [
            super_type for super_type in super_types
            if not isinstance(super_type, super_type_metatype)]
        if invalid_super_types:
            RAISE

        return metatype(name=self.name, superTypes=super_types)
    
    
    def _build_type_parameter(self, reader: DslReader):
        
        # TODO: rename type -> baseType
        
        type_expression = self._get_type_expression_attribute('type')
        0 and RAISE # check param?
        
        return self.metatype(name=self.name, type=type_expression)
        
    
    
    def _build_property(self, reader: DslReader):


        metatype = self.metatype
        type_expression = self._get_type_expression_attribute('type')

        0 and RAISE # check param
        if issubclass(metatype, metamodel.DataProperty):
            if not isinstance(type_expression, metamodel.DataTypeExpression):
                ko
        else:
            assert issubclass(metatype, metamodel.ObjectProperty)
            if not isinstance(type_expression, metamodel.ClassExpression):
                ko
                
        prop_kwargs = {
            'name': self.name,
            'type_': type_expression,
            'lower': self._get_non_negative_integer_attribute('lower'),
            'upper': self._get_optional_non_negative_integer_attribute('upper'),
            'isOrdered': self._get_optional_bool_attribute('isOrdered') or False}
        
        if metatype in [metamodel.DataProperty, metamodel.ReferenceProperty]:
            prop_kwargs['isUnique'] = self._get_optional_bool_attribute('isUnique') or False

            
            
        if metatype is metamodel.ReferenceProperty:
            prop_kwargs['oppositeLower'] = self._get_non_negative_integer_attribute('oppositeLower')
            prop_kwargs['oppositeUpper'] = self._get_optional_non_negative_integer_attribute('oppositeUpper')
 
        return metatype(**prop_kwargs)


# TODO: move
            
def cap_to_camel_case(cap_case: str):
    parts: list[str] = []
    cap = True
    for char in cap_case:
        if char.strip():
            if not cap:
                char = char.lower()
            else:
                cap = False
            parts.append(char)
        else:
            cap = True
    return ''.join(parts)



# TODO -> TOLATERDO



# pylint: disable=missing-docstring

from __future__ import annotations

import hashlib
import itertools
import json
import operator
import pathlib
import platform
import re
import shutil
import subprocess
import tempfile
import typing as T
import unicodedata

from attrs import define
from lxml import etree
from pnb.mcl.metamodel import standard as metamodel

THIS_DIR = pathlib.Path(__file__).parent

ELK_JS_DIR = THIS_DIR / 'uml_diagrams_js'
NODE_MODULES_DIR = ELK_JS_DIR / 'node_modules'

NODE_PATH = 'node'
NPM_PATH = 'npm'

def check_dependencies():
    
    if NODE_MODULES_DIR.is_dir():
        return

    if platform.system() == 'Windows':
        use_shell = True
    elif platform.system() == 'Linux':
        use_shell = False
    else:
        raise EnvironmentError(f'unknown system: {platform.system()}')

    p = subprocess.Popen(
        [NPM_PATH, 'install'],
        cwd=ELK_JS_DIR,
        shell=use_shell)
    
    if p.wait():
        shutil.rmtree(NODE_MODULES_DIR, ignore_errors=True)
        raise EnvironmentError('installation of node modules failed')
        

class XmlNamespace(str):

    def __new__(cls, value: str):
        namespace = super().__new__(cls, value)
        namespace.__term_prefix = f'{{{value}}}' if value else ''
        return namespace

    def term(self, name: str):
        return self.__term_prefix + name

    def __getattr__(self, name):
        return self.term(name)
    


SVG = XmlNamespace('http://www.w3.org/2000/svg')


NBSP = unicodedata.lookup('NBSP')

LinkGetter = T.Callable[[metamodel.Element, ], T.Optional[str]]

EPS = 1e-6



class Point:
    def __init__(self, x: float, y: float):
        self.x = x
        self.y = y
        
    def __add__(self, other):
        return Point(self.x + other.x, self.y + other.y)
        
    def __repr__(self):
        return f'<{self.x}, {self.y}>'
    
Font = T.Literal['normal', 'element']
FontSize = T.Literal['normal', 'large']

@define(eq=False)
class Text:
    text: str
    font: Font
    font_size: FontSize
    link: T.Optional[str]=None
    bold: bool=False
    italic: bool=False
    underlined: bool=False
    
    
Alignment = T.Literal['left', 'center']

LineStyle = T.Literal['solid', 'short_dash']

@define
class Cell:
    alignment: Alignment
    parts: list[Text] # TODO rename texts

    def iter_texts(self) -> T.Iterator[Text]:
        yield from self.parts

    def get_size(self, renderer):
        width = 0
        height = 0
        for text in self.parts:
            text_width, text_height = renderer.get_text_size(text)
            width += text_width
            height = max(height, text_height)
        return width, height
    
    def render(self, renderer, origo, width):
        renderer.draw_texts(
            origo,
            width,
            self.alignment,
            self.parts)
    

@define
class Row:
    cells: list[Cell]
    
    def iter_texts(self) -> T.Iterator[Text]:
        return itertools.chain.from_iterable(cell.iter_texts() for cell in self.cells)

    def get_cell_sizes(self, renderer):
        return [cell.get_size(renderer) for cell in self.cells]
    
class Section:
    
    def iter_texts(self) -> T.Iterator[Text]:
        raise NotImplementedError()
    
    def get_size(self, renderer):
        raise NotImplementedError()

@define
class Table(Section):
    rows: list[Row]
    
    col_spacing = 3
    
    def iter_texts(self) -> T.Iterator[Text]:
        return itertools.chain.from_iterable(row.iter_texts() for row in self.rows)
    
    def get_col_and_row_dimensions(self, renderer): # TODO: rename to "min"
        
        col_widths = []
        row_heights = []

        for row in self.rows:
            cell_sizes = row.get_cell_sizes(renderer)
            col_widths = [
                max(col_width if col_width else 0., cell_size[0] if cell_size else 0.)
                for (col_width, cell_size)
                in itertools.zip_longest(col_widths, cell_sizes)]
            if cell_sizes:
                row_height = max(cell_size[1] for cell_size in cell_sizes)
            else:
                row_height = 0.
            row_heights.append(row_height)
            
        return col_widths, row_heights
        
    
    def get_size(self, renderer):
        
        col_widths, row_heights = self.get_col_and_row_dimensions(renderer)

        if col_widths:
            width = sum(col_widths) + (len(col_widths)-1) * self.col_spacing
        else:
            width = 0.
            
        if row_heights:
            height = sum(row_heights)
        else:
            height = 0.
        
        return width, height
    
    def render(self, renderer, origo, width, owner_width):

        col_widths, row_heights = self.get_col_and_row_dimensions(renderer)
        if not col_widths or not row_heights:
            return
        
        assert sum(col_widths) <= width + EPS
        # Auto-expand last col to get full required width.
        col_widths[-1] = max(width - sum(col_widths[:-1]), 0.)
        
        y = origo.y
        
        for row, row_height in zip(self.rows, row_heights):
            x = origo.x
            for cell, col_width in zip(row.cells, col_widths):
                cell.render(renderer, Point(x, y), col_width)
                x += col_width + self.col_spacing
            y += row_height


            

@define
class Separator(Section):

    def iter_texts(self) -> T.Iterator[Text]:
        return ()
    
    def get_size(self, renderer):
        return 0., 1.
    
    def render(self, renderer, origo, width, owner_width):

        renderer.draw_line([
            Point(origo.x - 0.5*(owner_width-width), origo.y),
            Point(origo.x + 0.5*(owner_width+width), origo.y)])


class DiagramItem:

    def on_add(self, specification: DiagramSpecification) -> None:
        pass
    
    def iter_texts(self) -> T.Iterator[Text]:
        raise NotImplementedError()        

    def elk_source(self, item_id: str, specification: DiagramSpecification, renderer: Renderer) -> tuple[dict[str, str], dict, dict]:
        raise NotImplementedError()
        
    layout: T.Callable


class ItemLayout:
    pass




BackgroundColor = T.Literal['type', 'main_type', 'instance', 'background', 'foreground']



@define(eq=False)
class Box(DiagramItem):
    background_color: BackgroundColor
    padding: float
    frame: bool
    sections: list[Section]
    top_right_table: T.Optional[Table] = None

    def iter_texts(self) -> T.Iterator[Text]:
        for section in self.sections:
            yield from section.iter_texts()
        if self.top_right_table:
            yield from self.top_right_table.iter_texts()
            

    def iter_section_origos(self, renderer, origo):
        x = origo.x + self.padding
        y = origo.y + self.padding
        for section in self.sections:
            yield Point(x, y)
            section_width, section_height = section.get_size(renderer)
            y += section.get_size(renderer)[1] + self.padding

    def elk_source(self, item_id: str, specification: DiagramSpecification, renderer: Renderer):
        width = 0.
        height = 0.
        for section in self.sections:
            section_width, section_height = section.get_size(renderer)
            width = max(width, section_width)
            height += section_height
        width += 2*self.padding                       # left and right
        height += (len(self.sections)+1)*self.padding # top and bottom and between sections
        
        if self.top_right_table:
            tr_table_width, tr_table_height = self.top_right_table.get_size(renderer)
     
            tr_table_width += 2*self.padding
            tr_table_height += 2*self.padding
            
            # Ensure box is significantly wider than top-right table.
            width = max(width, tr_table_width+10*self.padding)
            height += tr_table_height
            

        connectors = max(
            specification.top_connectors_by_item.get(self, 0),
            specification.bottom_connectors_by_item.get(self, 0))
        min_width_for_connectors = connectors * 50
        width = max(width, min_width_for_connectors)

        return (
            {item_id: 'box_data'},
            [dict( id=item_id, width=width, height=height)],
            [])

    def layout(self, box_data):
        return BoxLayout(
            self,
            Point(float(box_data['x']), float(box_data['y'])),
            float(box_data['width']),
            float(box_data['height']))


@define
class BoxLayout(ItemLayout):
    box: Box
    position: Point
    width: width
    height: height
    
    def render(self, renderer):
        renderer.draw_box(
            self.position,
            self.width,
            self.height,
            frame=self.box.frame,
            background_color=self.box.background_color)
        
        section_width = self.width - 2*self.box.padding
        sections_origo = self.position

        if self.box.top_right_table:
            tr_table_width, tr_table_height = self.box.top_right_table.get_size(renderer)
            tr_table_width_padded = tr_table_width + 2*self.box.padding
            tr_table_height_padded = tr_table_height + 2*self.box.padding

            renderer.draw_line(
                points=[
                    Point(
                        self.position.x + self.width - tr_table_width_padded,
                        self.position.y),
                    Point(
                        self.position.x + self.width - tr_table_width_padded,
                        self.position.y + tr_table_height_padded),
                    Point(
                        self.position.x + self.width,
                        self.position.y + tr_table_height_padded)],
                line_style='short_dash')

            self.box.top_right_table.render(
                renderer,
                Point(
                    self.position.x + self.width - (tr_table_width + self.box.padding),
                    self.position.y + self.box.padding),
                tr_table_width,
                tr_table_width+2*self.box.padding)

            sections_origo = Point(sections_origo.x, sections_origo.y + tr_table_height_padded)

        for section, section_origo in zip(
                self.box.sections, self.box.iter_section_origos(renderer, sections_origo)):
            section.render(renderer, section_origo, section_width, self.width)



LineMarker = T.Literal['closed_arrow', 'filled_diamond', 'open_arrow']

@define(eq=False)
class Line(DiagramItem):
    source_id: str
    target_id: str
    label_box: T.Optional[Box] = None
    source_label: T.Optional[str] = None
    source_marker: T.Optional[LineMarker] = None
    target_label: T.Optional[str] = None
    target_marker: T.Optional[LineMarker] = None
    line_style: short_dtyle = 'solid'
    

    def on_add(self, specification: DiagramSpecification) -> None:
        source_item = specification.item_by_id.get(self.source_id)
        assert source_item
        specification.bottom_connectors_by_item[source_item] = specification.bottom_connectors_by_item.get(source_item, 0) + 1
        target_item = specification.item_by_id.get(self.target_id)
        assert target_item
        specification.top_connectors_by_item[target_item] = specification.top_connectors_by_item.get(target_item, 0) + 1
    
    def iter_texts(self) -> T.Iterator[Text]:
        
        
        # TODO: merge with elk source
        yield Text(
            text=self.source_label or '.',
            font='normal',
            font_size='normal')
        yield Text(
            text=self.target_label or '.',
            font='normal',
            font_size='normal')
        if self.label_box:
            yield from self.label_box.iter_texts()

        
        return [] # TODO

    def elk_source(self, item_id: str, specification: DiagramSpecification, renderer: Renderer):
        
        label_box_id = f'{item_id}-label-box'
        source_to_aux_edge_id = f'{item_id}-from'
        aux_to_target_edge_id = f'{item_id}-to'
        
        id_mapping = {
            label_box_id: 'label_box_data',
            source_to_aux_edge_id: 'source_to_aux_edge_data',
            aux_to_target_edge_id: 'aux_to_target_edge_data'}
        
        if self.label_box:
            label_box_id_mapping, label_box_children, label_box_edges = self.label_box.elk_source(
                label_box_id, specification, renderer)
            assert len(label_box_id_mapping) == 1
            assert label_box_id_mapping[label_box_id] == 'box_data'
            assert len(label_box_children) == 1
            assert not label_box_edges
            label_box_data = label_box_children[0]
        else:
            label_box_data = dict(id=label_box_id, width=1, height=1)

        source_to_aux_edge_data = dict(
            id=source_to_aux_edge_id,
            sources=[self.source_id],
            targets=[label_box_id])
        source_label_width, source_label_height = renderer.get_text_size(Text(
            text=self.source_label or '.',
            font='normal',
            font_size='normal'))
        source_to_aux_edge_data['labels'] = [{
            'text': '.',
            'layoutOptions': { 'edgeLabels.placement': 'TAIL'},
            'width': 1,
            'height': source_label_height}]

        aux_to_target_edge_data = dict(
            id=aux_to_target_edge_id,
            sources=[label_box_id],
            targets=[self.target_id])
        target_label_width, target_label_height = renderer.get_text_size(Text(
            text=self.target_label or '.',
            font='normal',
            font_size='normal'))
        aux_to_target_edge_data['labels'] = [{
            'text': '.',
            'layoutOptions': { 'edgeLabels.placement': 'HEAD'},
            'width': 1,
            'height': target_label_height}]

        return (
            id_mapping,
            [label_box_data],
            [source_to_aux_edge_data, aux_to_target_edge_data])
        
    def layout(self, label_box_data, source_to_aux_edge_data, aux_to_target_edge_data) -> LineLayout:

        # TODO: move below

        def point_from_data(data):
            return Point(float(data['x']), float(data['y']))

        def points_from_sections(data: dict):
            sections = data.get('sections')
            assert isinstance(sections, list)
            assert len(sections) == 1
            section = sections[0]
            points = [point_from_data(section.get('startPoint'))]
            bend_points_data = section.get('bendPoints')
            if bend_points_data:
                assert isinstance(bend_points_data, list)
                points.extend(point_from_data(point_data) for point_data in bend_points_data)
            points.append(point_from_data(section.get('endPoint')))
            return cleaned_points(points)

        def cleaned_points(points):
            return points

        if self.label_box:
            label_box_layout = self.label_box.layout(label_box_data)
        else:
            label_box_layout = None

        if self.source_label:
            source_label_position = (
                point_from_data(source_to_aux_edge_data['labels'][0]) + Point(5, 1.5))
        else:
            source_label_position = None

        if self.target_label:
            target_label_position = (
                point_from_data(aux_to_target_edge_data['labels'][0]) + Point(5, 1.5))
        else:
            target_label_position = None

        return LineLayout(
            self,
            label_box_layout=label_box_layout,
            source_to_aux_points=points_from_sections(source_to_aux_edge_data),
            aux_to_target_points = points_from_sections(aux_to_target_edge_data),
            source_label_position=source_label_position,
            target_label_position=target_label_position)


@define
class LineLayout(ItemLayout):
    line: Line
    label_box_layout: T.Optional[BoxLayout]
    source_to_aux_points: list[Point]
    aux_to_target_points: list[Point]
    source_label_position: T.Optional[Point]
    target_label_position: T.Optional[Point]

    def render(self, renderer):
        if self.line.label_box:
            assert self.label_box_layout
            self.label_box_layout.render(renderer)
            renderer.draw_line(
                self.source_to_aux_points,
                source_marker=self.line.source_marker,
                line_style=self.line.line_style) # TODO: clean
            renderer.draw_line(
                self.aux_to_target_points,
                target_marker=self.line.target_marker,
                line_style=self.line.line_style) # TODO: clean
        else:
            renderer.draw_line(
                self.source_to_aux_points + self.aux_to_target_points, # TODO: clean
                source_marker=self.line.source_marker,
                target_marker=self.line.target_marker,
                line_style=self.line.line_style)
            
        if self.line.source_label:
            assert self.source_label_position
            renderer.draw_texts(
                position=self.source_label_position,
                width=None,
                alignment='left',
                texts=[Text(text=self.line.source_label, font='normal', font_size='normal')])
            
        if self.line.target_label:
            assert self.target_label_position
            renderer.draw_texts(
                position=self.target_label_position,
                width=None,
                alignment='left',
                texts=[Text(text=self.line.target_label, font='normal', font_size='normal')])



class DiagramSpecification:

    def __init__(self) -> None:
        self.item_by_id: dict[str, DiagramItem] = {}
        self.id_by_item: dict[DiagramItem, str] = {}
        self.top_connectors_by_item: dict[DiagramItem, int] = {}
        self.bottom_connectors_by_item: dict[DiagramItem, int] = {}

    def add(self, item: DiagramItem) -> None:
        assert item not in self.id_by_item
        id_ = f'i{len(self.id_by_item)}'
        self.item_by_id[id_] = item
        self.id_by_item[item] = id_
        item.on_add(self)
        
    def iter_texts(self):
        return itertools.chain.from_iterable(item.iter_texts() for item in self.id_by_item)

    def get_hash(self) -> str:
        hash_obj = hashlib.sha256()
        hash_obj.update(repr(self.item_by_id).encode('utf-8'))
        return hash_obj.hexdigest()

    def make_layout(self, renderer: Renderer) -> DiagramLayout:
        children: list[dict] = []
        edges: list[dict] = []
        id_mapping_by_item_id: dict[str, dict[str, str]] = {}
        for item_id, item in self.item_by_id.items():
            item_id_mapping, item_children, item_edges = item.elk_source(item_id, self, renderer)
            id_mapping_by_item_id[item_id] = item_id_mapping
            children.extend(item_children)
            edges.extend(item_edges)

        elk_input = {
            'id': 'root',
            'layoutOptions': {
                'algorithm': 'layered',
                'elk.direction': 'DOWN'},
            'children': children,
            'edges': edges}

        with (ELK_JS_DIR / 'elk_source.json').open('w', encoding='utf-8') as elk_source_file:
            json.dump(elk_input, elk_source_file, indent=2)
            
        check_dependencies()

        cmd = [NODE_PATH, ELK_JS_DIR / 'run_elk.js']
        with subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE) as elk_process:
            elk_stdout, elk_stderr = elk_process.communicate()
        assert not elk_stderr, elk_stderr
        elk_output = json.loads(elk_stdout)

        data_by_id = {}
        for section in ('children', 'edges'):
            for data in elk_output.get(section, ()):
                data_id = data.get('id')
                assert data_id
                assert data_id not in data_by_id
                data_by_id[data_id] = data

        layouts: list[ItemLayout] = []

        for item_id, item in self.item_by_id.items():
            item_id_mapping = id_mapping_by_item_id[item_id]
            layout_kwargs = {}
            for data_id, data_name in item_id_mapping.items():
                data = data_by_id.pop(data_id, None)
                assert data
                assert data_name not in layout_kwargs
                layout_kwargs[data_name] = data
            layouts.append(item.layout(**layout_kwargs))

        return DiagramLayout(
            self,
            origo=Point(float(elk_output['x']), float(elk_output['y'])),
            width=float(elk_output['width']),
            height=float(elk_output['height']),
            layouts=layouts)

@define
class DiagramLayout:
    specification: DiagramSpecification
    origo: Point
    width: float
    height: float
    layouts: list[ItemLayout]

    def render(self, renderer: Renderer):
        for layout in self.layouts:
            layout.render(renderer)


class Renderer:

    @classmethod
    def get_result(cls,
            specification: DiagramSpecification,
            out: T.Optional[pathlib.Path]=None,
            cache_dir: T.Optional[pathlib.Path]=None):

        result = None

        if cache_dir:
            specification_hash = specification.get_hash()
            cached_diagram_path = cache_dir / f'{specification_hash}.{cls.extension}'
            if cached_diagram_path.is_file():
                result = cls.read_result(cached_diagram_path)

        if result is None:
            result = cls(specification).result
            if cache_dir:
                cls.write_result(cached_diagram_path, result)
        
        if out:
            cls.write_result(out, result)

        return result

    @classmethod
    def write_result(self, path: pathlib.Path, result: object):
        raise NotImplementedError()

    def __init__(self, specification):
        self._register_texts(specification)
        layout = specification.make_layout(self)
        self._init_layout(layout)
        layout.render(self)

    @property
    def result(self):
        raise NotImplementedError()

    def _register_texts(self, specification: DiagramSpecification):
        pass

    def _init_layout(self, layout: DiagramLayout):
        pass



class SvgRenderer(Renderer):

    extension = 'svg'
    marker_size = 10
    marker_defs=f'''
      <defs xmlns="{SVG}">
        <marker
            id="uml-marker-closed_arrow"
            viewBox="-2 -2 {marker_size+4} {marker_size+4}"
            refX="{marker_size}"
            refY="{0.5*marker_size}"
            markerWidth="{marker_size+4}"
            markerHeight="{marker_size+4}"
            orient="auto-start-reverse">
          <path
              d="M 0 0 L {marker_size} {0.5*marker_size} L 0 {marker_size} Z"
              stroke-linejoin="round"
              stroke-linecap="round"
              class="uml-foreground-stroke uml-background_color-background"
          />
        </marker>
        <marker
            id="uml-marker-open_arrow"
            viewBox="-2 -2 {marker_size+4} {marker_size+4}"
            refX="{marker_size}"
            refY="{0.5*marker_size}"
            markerWidth="{marker_size+4}"
            markerHeight="{marker_size+4}"
            orient="auto-start-reverse">
          <path
              d="M 0 0 L {marker_size} {0.5*marker_size} L 0 {marker_size}"
              stroke-linejoin="round"
              stroke-linecap="round"
              fill="none"
              class="uml-foreground-stroke"
          />
        </marker>
        <marker
            id="uml-marker-filled_diamond"
            viewBox="-2 -2 {marker_size+4} {marker_size+4}"
            refX="{marker_size}"
            refY="{0.5*marker_size}"
            markerWidth="{marker_size+4}"
            markerHeight="{marker_size+4}"
            orient="auto-start-reverse">
          <path
              d="M 0 {0.5*marker_size} L      {0.5*marker_size} {0.2*marker_size}        {marker_size} {0.5*marker_size}      {0.5*marker_size} {0.8*marker_size} Z       "
              stroke-linejoin="round"
              stroke-linecap="round"
              class="uml-foreground-stroke uml-background_color-foreground"
          />
        </marker>
      </defs>'''

    @classmethod
    def write_result(cls, path: pathlib.Path, result):
        path.parent.mkdir(parents=True, exist_ok=True)
        path.write_bytes(etree.tostring(result, encoding='utf-8'))

    @classmethod
    def read_result(cls, path):
        return etree.parse(path).getroot()

    @property
    def result(self):
        return self.svg

    def _init_layout(self, layout: DiagramLayout):

        self.svg = etree.Element(SVG.svg)

        x0 = layout.origo.x
        y0 = layout.origo.y
        self.svg.attrib['width'] = f'{layout.width}px'
        self.svg.attrib['height'] = f'{layout.height}px'
        self.svg.attrib['viewBox'] = f'{x0} {y0} {x0+layout.width} {y0+layout.height}'

        # TODO: calculate style
        etree.SubElement(self.svg, SVG.style).text='''
            .uml-foreground-stroke {
                stroke: black;
                stroke-width: 1px;
            }
            
            .uml-foreground-fill {
                fill: black;
            }
            
            .uml-background_color-type {
                fill: #ffffcc;
            }
            
            .uml-background_color-main_type {
                fill: #ffff66;
            }
            
            .uml-background_color-instance {
                fill: blue;
            }
            
            .uml-background_color-background {
                fill: white;
            }
            
            .uml-background_color-foreground {
                fill: black;
            }
            
            .uml-font-normal {
                font-family: Arial;
            }
            
            .uml-font-element {
                font-family: Arial;
            }
            
            .uml-fontsize-normal {
                font-size: 10px;
            }

            .uml-fontsize-large {
                font-size: 12px;
            }
            
            .uml-source_marker-closed_arrow {
                marker-start: url(#uml-marker-closed_arrow);
            }
            
            .uml-source_marker-open_arrow {
                marker-start: url(#uml-marker-open_arrow);
            }
            
            .uml-source_marker-filled_diamond {
                marker-start: url(#uml-marker-filled_diamond);
            }

            .uml-target_marker-closed_arrow {
                marker-end: url(#uml-marker-closed_arrow);
            }
            
            .uml-target_marker-open_arrow {
                marker-end: url(#uml-marker-open_arrow);
            }
            
            .uml-target_marker-filled_diamond {
                marker-end: url(#uml-marker-filled_diamond);
            }

            '''

        self.svg.append(etree.XML(self.marker_defs))

    def get_text_size(self, text: Text):
        


        if 1:
            
            #text.font = 'element'
            #text.bold = False
           # text.italic = False
            text.font_size = 'large'
            

        
        
        scale = {
            'normal': 10.,
            'large': 12.}[text.font_size] # check with css
            
            
            
            
        if 1:
            width = sum(
                WIDTH_BY_CHAR[char] for char in text.text)
            
            if text.bold:
                width *= 1.1
            if text.italic:
                width *= 1.1
            
            return width*(scale/12), scale*1.15

        return scale * 0.52 * len(text.text), scale*1.15

    def draw_box(self, position: Point, width: float, height: float, frame: bool, background_color: BackgroundColor):
        etree.SubElement(self.svg, SVG.rect, {
            'class': ' '.join([
                'uml-background_color-' + background_color,
                'uml-foreground-stroke' if frame else '']),
            'x': str(position.x),
            'y': str(position.y),
            'width': str(width),
            'height': str(height)})

    def draw_line(self, points: list[Point], source_marker=None, target_marker=None, line_style: LineStyle='solid'):
        classes = ['uml-foreground-stroke']
        if source_marker:
            classes.append(f'uml-source_marker-{source_marker}')
        if target_marker:
            classes.append(f'uml-target_marker-{target_marker}')
        polyline_element = etree.SubElement(self.svg, SVG.polyline, {
            'class': ' '.join(classes),
            'points': ' '.join(f'{pt.x},{pt.y}' for pt in points),
            'fill': 'none'})
        
        if line_style == 'short_dash':
            polyline_element.attrib['stroke-dasharray'] = '4 2'

        else:
            assert line_style == 'solid'
        

    def draw_texts(self, position: Point, width: double, alignment: alignment, texts: list[Text]):
        
        if not texts:
            return
        
        text_element = etree.SubElement(self.svg, SVG.text)

        height = 0
        
        for text in texts:
            if text.link:
                tspan_parent_element = etree.SubElement(text_element, SVG.a, {'href': text.link})
            else:
                tspan_parent_element = text_element
                
            if text.font_size == 'normal':
                height = max(height, 10.)
            else:
                height = max(height, 12.)

            tspan_attrib = {
                'class': ' '.join([
                    'uml-foreground-fill',
                    f'uml-font-{text.font}',
                    f'uml-fontsize-{text.font_size}',
                    ])}

            if text.bold:
                tspan_attrib['font-weight'] = 'bold'
            if text.italic:
                tspan_attrib['font-style'] = 'italic'
            if text.underlined:
                tspan_attrib['text-decoration'] = 'underline'

            tspan_element = etree.SubElement(tspan_parent_element, SVG.tspan, tspan_attrib)
            tspan_element.text = text.text.replace(' ', NBSP)

        if alignment == 'left':
            text_element.attrib['x'] = str(position.x)
        else:
            assert alignment == 'center'
            text_element.attrib['x'] = str(position.x + 0.5*width)
            text_element.attrib['text-anchor'] = 'middle'
            
        # Make layout agnostic to any whitespace between <tspan> elements, e.g., to line breaks
        # from prettyprint.
        text_element.attrib['font-size'] = '0px'
            
        BASELINE_FACTOR = 0.87 # TODO: calculate???    
        
        text_element.attrib['y'] = str(position.y + BASELINE_FACTOR * height)
        
        if False:
        
        
            etree.SubElement(self.svg, SVG.rect,
             {
                 'x': str(position.x - 10),
                 'y': str(position.y),
                 'width': '5',
                 'height': str(height),
                 'stroke': 'black',
                 'fill': 'green',
                 
                 
                 })


class LaTeXRenderer(Renderer):
    
    extension = 'tex'
    
    global_scale = 1
    
    _width_by_tex_text: dict[str, float] = {}
    
    @classmethod
    def write_result(cls, path: pathlib.Path, result):
        path.parent.mkdir(parents=True, exist_ok=True)
        path.write_text(result, encoding='utf-8') # TODO: encoding?

    @classmethod
    def read_result(cls, path):
        return path.read_text(encoding='utf-8') # TODO: encoding?
    
    
    
    @property
    def result(self):
        if self.code is None:
            self.code_parts.append(r'\end{tikzpicture}')
            self.code = '\n'.join(self.code_parts)
        return self.code
    
    
    def _register_texts(self, specification: DiagramSpecification):
        
        to_calculate: set[str] = set()
        for text in specification.iter_texts():
            tex_text = self.text_to_tex(text)
            if tex_text not in self._width_by_tex_text:
                to_calculate.add(tex_text)
                
        if to_calculate:
            self._width_by_tex_text.update(self._calculate_text_widths(to_calculate))
            
            
    _text_calculation_start = r'''
        \def\sphinxdocclass{book}
        \documentclass[10pt,english]{scrbook}
        \usepackage{printlen}
        \usepackage{hyperref}
        \usepackage{tikz}
        
        % This provides a bold+italic style for sf!
        \usepackage{lmodern}
        
        
        
        \usetikzlibrary{patterns} 
        \begin{document}
        \newlength{\gnat}'''
    
    _text_calculation_end = r'''
        \end{document}'''
    
    
            
                
    def _calculate_text_widths(self, tex_texts: set[str]):
        
        latex_code_parts = [
            self._text_calculation_start]
        
        for nr, text in enumerate(tex_texts):
            latex_code_parts += [
                fr'\settowidth{{\gnat}}{{{text}}}',
                fr'\message{{--WIDTH--\printlength{{\gnat}}--}}']
                
        latex_code_parts.append(self._text_calculation_end)
        latex_code = '\n'.join(latex_code_parts)
        
        
        with tempfile.TemporaryDirectory() as tmp_dir_name:
            tmp_dir = pathlib.Path(tmp_dir_name)
            tex_path = tmp_dir / 'calculate_widths.tex'
            tex_path.write_text(latex_code, encoding='utf-8')
            # print("CALL", tex_path.name)
            tex_result = subprocess.check_output(
                ['pdflatex', tex_path.name],
                cwd=tmp_dir)

        width_matches = list(re.finditer(
            rb'--WIDTH--\\def pt\{pt\}(?P<width>[0-9]*\.?[0-9]*)pt--',
            tex_result))
        
        assert len(width_matches) == len(tex_texts)
        
        result: dict[str, float] = {}
        
        for tex_text, width_match in zip(tex_texts, width_matches):
            width_string = width_match['width']
            width = float(width_string)
            result[tex_text] = width
        
        return result
                
                
                

    
    
    def _init_layout(self, layout: DiagramLayout):
        
       # for text in layout.
        
        
        
        
        
        
        
        
        self.scale = 0.02
        
        self.height = layout.height
        
        
        self.code = None
        
        self.code_parts: list[str] = [
            r'\begin{tikzpicture}']
        
        return
        

    
    
    def get_text_size(self, text: Text):
        

        
        width = self._width_by_tex_text[self.text_to_tex(text)]
        
  
            
            
        


            
            

        
        
        scale = {
            'normal': 10.,
            'large': 12.}[text.font_size] # check with css
            
            
            
            
        if 1:
           # width = sum(
           #     WIDTH_BY_CHAR[char] for char in text.text)
            
           # if text.bold:
           #     width *= 1.1
           # if text.italic:
           #     width *= 1.1
            
            return 1.88*width/self.global_scale, scale*1.55/self.global_scale

        return scale * 0.52 * len(text.text), scale*1.15
    
    
    def draw_box(self, position: Point, width: float, height: float, frame: bool, background_color: BackgroundColor):
        
        opp_position = position + Point(width, height)
        fill = {
            'type': 'lime',
            'main_type': 'yellow',
            'background': 'white'}[background_color]
        command = r'\draw' if frame else r'\fill'
        self.code_parts.append(fr'{command}[fill={fill}] {self.point_to_tex(position)} rectangle {self.point_to_tex(opp_position)};')

        
    def point_to_tex(self, point):
        return f'({self.global_scale*self.scale*point.x},{self.global_scale*self.scale*(self.height-point.y)})'
    
    def length_to_tex(self, length):
        print('++++++++++++++++', length, self.global_scale*self.scale*length)
        return self.scale*length

    def draw_line(self, points: list[Point], source_marker=None, target_marker=None, line_style: LineStyle='solid'):

        line_parts = [r'\draw']
        
        if source_marker or target_marker:
            markercode = '['
            if source_marker:
                markercode += {
                    'filled_diamond': '{Diamond}',
                    'closed_arrow': '{Triangle[open]}'}[source_marker]
            markercode += '-'
            if target_marker:
                markercode += {'open_arrow': '{Stealth}'}[target_marker]
            markercode += ']'
            line_parts.append(markercode)

        if line_style == 'short_dash':
            line_parts.append('[dashed]')

        line_parts.append(self.point_to_tex(points[0]))
        for point in points[1:]:
            line_parts.append(f'--{self.point_to_tex(point)}')
        line_parts.append(';')
        line = ''.join(line_parts)
        self.code_parts.append(line)
        
 
            
    font_to_tex = {
        'normal': r'\textsf',#r'\textsf',
        'element': r'\textsf',#r'\textsf'}
        }
    
    font_size_to_tex = {
        'normal': r'\footnotesize', #r'\normalsize',
        'large': r'\small', #r'\large',
        }
    
    bold_to_tex = {
        True: r'\textbf',
        False: ''
        }
    
    italic_to_tex = {
        True: r'\emph',
        False: ''
    }
    
    
            
    def text_to_tex(self, text):
            # text: str
            # font: Font
            # font_size: FontSize
            # link: T.Optional[str]=None
            # bold: bool=False
            # italic: bool=False
            # underlined: bool=False
            
            
            
        # strut
        


        
        tex_parts = [
            '{',
            self.font_to_tex[text.font],
                        '{',
            self.font_size_to_tex[text.font_size],
                      '{',
            self.italic_to_tex[text.italic],
             '{',
            
            self.bold_to_tex[text.bold],

           '{',
            self.tex_escape(text.text),
            '}}}}}']
        
        return ''.join(tex_parts)
    
    
    
    
    tex_escape_map = {
        '|': r'$\vert$',
        '<': r'$<$',
        '>': r'$>$'}

    
    def tex_escape(self, text: str) -> str:
        return ''.join(self.tex_escape_map.get(char, char) for char in text)


    def draw_texts(self, position: Point, width: double, alignment: Alignment, texts: list[Text]):
        
        font_sizes = set(text.font_size for text in texts)
        font_sizes = set({'large': 12, 'normal': 10}[font_size] for font_size in font_sizes)
        font_size = max(font_sizes)
        tex_text = ''.join(self.text_to_tex(text) for text in texts)
    
        position += Point(0, 0.85*font_size)
        if alignment == 'left':
            anchor = '[anchor=west]'
        else:
            anchor = ''
            position += Point(0.5*width, 0)
            
        self.code_parts.append(
            fr'\node{anchor} at {self.point_to_tex(position)} {{\strut{tex_text}}};')


class DiagramSpecificationBuilder:

    def __init__(self, *, get_link: T.Optional[LinkGetter]=None):
        self.diagram_specification = DiagramSpecification()
        self.get_link = get_link or (lambda element: None)
        self.node_by_element: dict[metamodel.Element, Node] = {}


    def get_node_id(self, element: T.Optional[metamodel.Element]) -> str:
        node = self.node_by_element.get(element)
        if not node:
            print(element)
            TODO
        id_ = self.diagram_specification.id_by_item.get(node)
        if not id_:
            TODO
        return id_


    def add_type_expression_box(
            self, type_expr: metamodel.TypeExpression, is_main_element: bool) -> None:
        header_rows = []
        if isinstance(type_expr, metamodel.Enumeration):
            header_rows.append(self.stereotype_row('enumeration'))
        elif isinstance(type_expr, metamodel.PrimitiveType):
            header_rows.append(self.stereotype_row('primitive'))
        elif isinstance(type_expr, metamodel.DataType):
            header_rows.append(self.stereotype_row('datatype'))
        elif isinstance(type_expr, metamodel.UnionType):
            header_rows.append(self.stereotype_row('union'))
        elif isinstance(type_expr, metamodel.BoundType):
            header_rows.append(self.stereotype_row('bind'))

        if isinstance(type_expr, metamodel.Type):
            header_rows.append(self.title_row(
                type_expr.name,
                font='element',
                bold=True,
                italic=type_expr.isAbstract,
                link=self.get_link(type_expr)))
        sections: list[Section] = [Table(rows=header_rows)]

        top_right_table = None
        if isinstance(type_expr, (metamodel.Class, metamodel.AggregatedDataType)):
            template_parameter_rows = []
            for template_parameter in type_expr.ownedParameters:
                cell_1 = Cell(
                    alignment='left',
                    parts=[
                        Text(
                            text=template_parameter.name,
                            font='element',
                            font_size='normal',
                            link=self.get_link(template_parameter)),
                        Text(
                            text=' :',
                            font='element',
                            font_size='normal')])
                cell_2 = Cell(
                    alignment='left',
                    parts=self.get_type_reference_texts(template_parameter.type, font_size='normal'))
                template_parameter_rows.append(Row([cell_1, cell_2]))
            if template_parameter_rows:
                top_right_table = Table(template_parameter_rows)

        if is_main_element and isinstance(type_expr, metamodel.Type):
            # Add Table with data properties to sections.
            property_rows = []
            for prop in type_expr.ownedAttributes:
                if not isinstance(prop, metamodel.DataProperty):
                    continue
                assert prop.name
                cell_1 = Cell(
                    alignment='left',
                    parts=[
                        Text(
                            text=prop.name,
                            font='element',
                            font_size='normal',
                            link=self.get_link(prop)),
                        Text(
                            text=' :',
                            font='element',
                            font_size='normal')])
                cell_2 = Cell(
                    alignment='left',
                    parts=self.get_type_reference_texts(prop.type, font_size='normal'))
                more_texts: list[str] = []
                multiplicity_string = multiplicity_label(prop.lower, prop.upper)
                if multiplicity_string:
                    more_texts.append(f'[{multiplicity_string}]')
                modifiers = []
                if prop.isOrdered:
                    modifiers.append('ordered')
                if prop.isUnique:
                    modifiers.append('unique')
                if modifiers:
                    more_texts.append(f'{{{", ".join(modifiers)}}}')
                if more_texts:
                    cell_2.parts.append(
                        Text(
                            text=' ' + ' '.join(more_texts),
                            font='element',
                            font_size='normal'))
                property_rows.append(Row([cell_1, cell_2]))
            if property_rows:
                sections.append(Separator())
                sections.append(Table(property_rows))
                
                
                
                
        if is_main_element and isinstance(type_expr, metamodel.Class):
            
          extensions =   type_expr._extensions
          
          for extension in extensions:

            # Add Table with data properties to sections.
            property_rows = []
            for prop in extension.ownedAttributes:
                if not isinstance(prop, metamodel.DataProperty):
                    continue
                assert prop.name
                cell_1 = Cell(
                    alignment='left',
                    parts=[
                        Text(
                            text=prop.name,
                            font='element',
                            font_size='normal',
                            link=self.get_link(prop)),
                        Text(
                            text=' :',
                            font='element',
                            font_size='normal')])
                cell_2 = Cell(
                    alignment='left',
                    parts=self.get_type_reference_texts(prop.type, font_size='normal'))
                more_texts: list[str] = []
                multiplicity_string = multiplicity_label(prop.lower, prop.upper)
                if multiplicity_string:
                    more_texts.append(f'[{multiplicity_string}]')
                modifiers = []
                if prop.isOrdered:
                    modifiers.append('ordered')
                if prop.isUnique:
                    modifiers.append('unique')
                if modifiers:
                    more_texts.append(f'{{{", ".join(modifiers)}}}')
                if more_texts:
                    cell_2.parts.append(
                        Text(
                            text=' ' + ' '.join(more_texts),
                            font='element',
                            font_size='normal'))
                property_rows.append(Row([cell_1, cell_2]))
            if property_rows:
                
                source_element = extension
                while source_element.owner:
                    source_element = source_element.owner
                
                sections.append(Separator())
                sections.append(Table([
                    Row([Cell(
                        alignment='left',
                        parts=[
                            Text(
                                text='from ',
                                font='element',
                                font_size='normal',
                                italic=True),
                            Text(
                                text=source_element.name+':',
                                font='element',
                                font_size='normal',
                                link=self.get_link(source_element),
                                italic=True)

                            
                            
                            
                            ])])]))
                
                sections.append(Table(property_rows))
                

                
                
        if is_main_element and isinstance(type_expr, metamodel.Enumeration):
            # Add Table with enumeration literals.
            literal_rows = []
            for literal in type_expr.ownedLiterals:
                literal_rows.append(self.texts_to_row(
                    alignment='left',
                    inlines=[Text(
                        text=literal.name,
                        font='normal',
                        font_size='normal',
                        link=self.get_link(literal))]))
            if literal_rows:
                sections.append(Separator())
                sections.append(Table(literal_rows))
                
        if isinstance(type_expr, metamodel.BoundType):
            bindung_rows = []
            for binding in type_expr.bindings:
                cell_1 = Cell(
                    alignment='left',
                    parts=[
                        Text(
                            text=binding.parameter.name,
                            font='element',
                            font_size='normal',
                            link=self.get_link(binding.parameter)),
                        Text(
                            text=' →',
                            font='element',
                            font_size='normal')])
                cell_2 = Cell(
                    alignment='left',
                    parts=self.get_type_reference_texts(binding.type, font_size='normal'))
                bindung_rows.append(Row([cell_1, cell_2]))
            if bindung_rows:
                sections.append(Separator())
                sections.append(Table(bindung_rows))

        box = Box(
            background_color='main_type' if is_main_element else 'type',
            frame=True,
            padding=5,
            sections=sections,
            top_right_table=top_right_table)
        self.node_by_element[type_expr] = box
        self.diagram_specification.add(box)


    def add_generalization_line(self, sub_type: metamodel.Type, super_type: metamodel.Type):
        line = Line(
            source_id=self.get_node_id(super_type),
            target_id=self.get_node_id(sub_type),
            source_marker='closed_arrow')
        self.diagram_specification.add(line)
        
        
    def add_binding_line(self, referrer: metamodel.TypeExpression, base: metamodel.TypeExpression):
        line = Line(
            source_id=self.get_node_id(referrer),
            target_id=self.get_node_id(base),
            target_marker='open_arrow',
            line_style='short_dash')
        self.diagram_specification.add(line)



    def add_object_property_line(self, prop: metamodel.ObjectProperty) -> None:
        
        label_rows = [self.texts_to_row(
            alignment='center',
            inlines=[Text(
                text=prop.name,
                font='element',
                link=self.get_link(prop),
                font_size='normal')])]

        label_box = Box(
            background_color='background',
            padding=2.,
            frame=False,
            sections=[Table(rows=label_rows)])

        line = Line(
            source_id=self.get_node_id(prop.owner),
            target_id=self.get_node_id(prop.type),
            label_box=label_box,
            source_label=multiplicity_label(prop.oppositeLower, prop.oppositeUpper) or None,
            source_marker='filled_diamond' if isinstance(prop, metamodel.CompositionProperty) else None,
            target_label=multiplicity_label(prop.lower, prop.upper) or None,
            target_marker='open_arrow')
        self.diagram_specification.add(line)


    def stereotype_row(self, stereotype: str) -> Row:
        text = Text(
            text=f'<<{stereotype}>>',
            font='normal',
            font_size='normal')
        return self.texts_to_row('center', [text])


    def title_row(self, title: str, *, font: Font, bold: bool=False, italic: bool=False, link: T.Optional[str]=None) -> Row:
        text = Text(
            text=title,
            font=font,
            link=link,
            font_size='large',
            bold=bold,
            italic=italic)
        return self.texts_to_row('center', [text])


    def texts_to_row(self, alignment: Alignment, inlines: list[Text]) -> Row:
        return Row([Cell(alignment, inlines)])
    
    def get_type_reference_texts(
            self, type_reference: metamodel.TypeExpression, font_size: FontSize):
        if isinstance(type_reference, (metamodel.Type, metamodel.TypeParameter)):
            return [Text(
                text=type_reference.name,
                font='element',
                font_size=font_size,
                link=self.get_link(type_reference))]
        elif isinstance(type_reference, metamodel.UnionType):
            texts = []
            for base_type in type_reference.bases:
                if texts:
                    texts.append(Text(
                        text=' | ',
                        font='element',
                        font_size=font_size))
                texts.extend(self.get_type_reference_texts(base_type, font_size=font_size))
            return texts

        elif isinstance(type_reference, metamodel.BoundType):
            texts = [
                Text(
                    text='<',
                    font='element',
                    font_size=font_size),
                Text(
                    text=type_reference.base.name,
                    font='element',
                    font_size=font_size,
                    link=self.get_link(type_reference.base)),
                Text(
                    text=' with ',
                    font='element',
                    font_size=font_size)]
            
            binding_texts: list[Text] = []
            for binding in type_reference.bindings:
                if binding_texts:
                    texts.append(
                        Text(
                            text=', ',
                            font='element',
                            font_size=font_size))
                texts += [
                    Text(
                        text=binding.parameter.name,
                        font='element',
                        font_size=font_size,
                        link=self.get_link(binding.parameter)),
                    Text(
                        text=' → ',
                        font='element',
                        font_size=font_size)]
                texts += self.get_type_reference_texts(binding.type, font_size=font_size)
            texts.extend(binding_texts)
            texts.append(
                Text(
                    text='>',
                    font='element',
                    font_size=font_size))

            return texts
            
        else:
            TODO
            
            
            



def make_diagram_specification(
        elements: T.Iterable[metamodel.Type], *,
        usages_by_class: dict[metamodel.Class, list[metamodel.ObjectProperty]],
        get_link: T.Optional[LinkGetter]=None):

    main_elements: set[metamodel.Type] = set()
    hierarchy_types: set[metamodel.Type] = set()

    for element in elements:
        if element in main_elements:
            continue
        main_elements.add(element)
        assert isinstance(element, metamodel.Type)
        hierarchy_types.add(element)
        hierarchy_types |= element.allSuperTypes

    type_expressions: set[metamodel.TypeExpression] = set(hierarchy_types)
    object_properties = set()

    sub_types_by_type: dict[metamodel.Type, set[metamodel.Type]] = {}

    for type_ in hierarchy_types:
        sub_types_by_type[type_] = set(type_.subTypes).intersection(hierarchy_types)

        for prop in type_.ownedAttributes:
            if isinstance(prop, metamodel.ObjectProperty):
                object_properties.add(prop)
                type_expressions.add(prop.type)
                
            for prop in sorted(
                    usages_by_class.get(type_, []),
                    key=lambda prop: prop.qualifiedName):
                
                if not isinstance(prop.type, metamodel.Class):
                    # Can we handle BoundClasses?
                    continue

                object_properties.add(prop)
                type_expressions.add(prop.owner)
        # TODO: opp props
      
      
        
    type_expressions_to_handle = type_expressions
    handled_type_expressions: set[metamodel.TypeExpression] = set()
    
    union_bases: list[tuple[metamodel.UnionType, metamodel.TypeExpression]] = []
    binding_bases: list[tuple[metamodel.BoundType, metamodel.Type]] = []
    
    while type_expressions_to_handle:
        type_expr = type_expressions_to_handle.pop()
        handled_type_expressions.add(type_expr)
        if isinstance(type_expr, metamodel.UnionType):
            for base in type_expr.bases:
                if base not in handled_type_expressions:
                    type_expressions_to_handle.add(base)
                union_bases.append((type_expr, base))
        elif isinstance(type_expr, metamodel.BoundType):
            if type_expr.base not in handled_type_expressions:
                type_expressions_to_handle.add(type_expr.base)
            binding_bases.append((type_expr, type_expr.base))
    type_expressions = handled_type_expressions
                    
        
        
    
    
    
    

    symbol_by_element = {}
    
    builder = DiagramSpecificationBuilder(get_link=get_link)

    for type_expr in sorted(type_expressions, key=repr):
        symbol_by_element[type_expr] = builder.add_type_expression_box(type_expr, type_expr in main_elements)
        

    for type_ in sorted(hierarchy_types, key=repr):
        for sub_type in sorted(sub_types_by_type[type_], key=lambda type_: type_.qualifiedName):
            builder.add_generalization_line(sub_type, type_)
    
    for prop in sorted(object_properties, key=lambda prop: prop.qualifiedName):
        builder.add_object_property_line(prop)
 
 
    for referrer, base in sorted(binding_bases, key=repr):
        builder.add_binding_line(referrer, base)
        
   # union_bases: list[tuple[metamodel.UnionType, metamodel.TypeExpression]] = []
   # binding_bases: list[tuple[metamodel.BoundType, metamodel.Type]] = []

    return builder.diagram_specification




GraphicsFormat = T.Literal['svg', 'tex']

def make_diagram(
        elements: T.Iterable[metamodel.Type], 
        format: GraphicsFormat, *,
        out: T.Optional[pathlib.Path]=None,
        get_link: T.Optional[LinkGetter]=None,
        cache_dir: T.Optional[pathlib.Path]=None,
        usages_by_class: T.Optional[dict[metamodel.Class, list[metamodel.ObjectProperty]]]):
    
    #if elements[0].name != 'Composition':
     ############################################################################################   return

    diagram_specification = make_diagram_specification(elements, get_link=get_link, usages_by_class=usages_by_class or {})
    renderer_cls = {
        'svg': SvgRenderer,
        'tex': LaTeXRenderer}[format]
        
    return renderer_cls.get_result(diagram_specification, out=out, cache_dir=cache_dir)


def sort_key(element: metamodel.Type):
    return element.qualifiedName



def multiplicity_label(lower: int, upper: T.Optional[int]):
    if upper is None:
        if lower == 0:
            return '*'
        else:
            return f'{lower}..*'
    elif lower == upper:
        if lower == 1:
            return ''
        else:
            return str(lower)
    else:
        return f'{lower}..{upper}'
    
    













WIDTH_BY_CHAR = { 
   "A": 8.00390625,
   "B": 8.00390625,
   "C": 8.666015625,
   "D": 8.666015625,
   "E": 8.00390625,
   "F": 7.330078125,
   "G": 9.333984375,
   "H": 8.666015625,
   "I": 3.333984375,
   "J": 6,
   "K": 8.00390625,
   "L": 6.673828125,
   "M": 9.99609375,
   "N": 8.666015625,
   "O": 9.333984375,
   "P": 8.00390625,
   "Q": 9.333984375,
   "R": 8.666015625,
   "S": 8.00390625,
   "T": 7.330078125,
   "U": 8.666015625,
   "V": 8.00390625,
   "W": 11.326171875,
   "X": 8.00390625,
   "Y": 8.00390625,
   "Z": 7.330078125,
   "a": 6.673828125,
   "b": 6.673828125,
   "c": 6,
   "d": 6.673828125,
   "e": 6.673828125,
   "f": 3.333984375,
   "g": 6.673828125,
   "h": 6.673828125,
   "i": 2.666015625,
   "j": 2.666015625,
   "k": 6,
   "l": 2.666015625,
   "m": 9.99609375,
   "n": 6.673828125,
   "o": 6.673828125,
   "p": 6.673828125,
   "q": 6.673828125,
   "r": 3.99609375,
   "s": 6,
   "t": 3.333984375,
   "u": 6.673828125,
   "v": 6,
   "w": 8.666015625,
   "x": 6,
   "y": 6,
   "z": 6,
   "0": 6.673828125,
   "1": 6.673828125,
   "2": 6.673828125,
   "3": 6.673828125,
   "4": 6.673828125,
   "5": 6.673828125,
   "6": 6.673828125,
   "7": 6.673828125,
   "8": 6.673828125,
   "9": 6.673828125,
   "_": 6.673828125,
   "<": 7.0078125,
   ">": 7.0078125,
   "[": 3.333984375,
   "]": 3.333984375,
   ".": 3.333984375,
   " ": 3.333984375,
   NBSP: 3.333984375,
   ' ': 3.333984375,
   ':': 2.5,
   '*': 6,
   '{': 4,
   '}': 4,
   ',': 3.5,
   '|': 2.5,
   '→': 11.,
   '\\': 0, # TODO
}